r"""Database operations on the stp_custom_role table."""

# Standard library
from typing import Iterable, Literal, Sequence, overload

# Third party
import pandas as pd
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

# Local
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.database.core import Session
from streamlit_passwordless.database.core import commit as db_commit
from streamlit_passwordless.database.models import CustomRole


@overload
def get_all_custom_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[False] = False,
    index_col: str | list[str] | None = 'role_id',
) -> Sequence[CustomRole]: ...


@overload
def get_all_custom_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[True] = True,
    index_col: str | list[str] | None = 'role_id',
) -> pd.DataFrame: ...


def get_all_custom_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: bool = False,
    index_col: str | list[str] | None = 'role_id',
) -> pd.DataFrame | Sequence[CustomRole]:
    r"""Get all custom roles from the database.

    Parameters
    ----------
    session : Session
        An active database session.

    skip : int, default 0
        The number of custom roles to skip from the beginning of the result set.
        The custom roles are order by their rank.

    limit : int or None, default None
        Limit the result to desired number of custom roles. If None all custom roles are returned.

    as_df : bool, default False
        If True return the custom roles as a :class:`pandas.DataFrame` instead of a sequence
        of :class:`streamlit_passwordless.db.models.CustomRole`.

    index_col : str or list[str] or None, default 'role_id'
        The column from the table stp_custom_role to use as the index column of the
        :class:`pandas.DataFrame` if `as_df` is True. Specify a list of column names
        to create a multiindex. If None no column is added to the index.

    Returns
    -------
    pandas.DataFrame or Sequence[streamlit_passwordless.db.models.CustomRole]
        The custom roles from the database.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the custom roles from the database.
    """

    query = select(CustomRole).order_by(CustomRole.rank).offset(skip).limit(limit)

    try:
        if as_df:
            return pd.read_sql_query(sql=query, con=session.bind, index_col=index_col)  # type: ignore
        else:
            return session.scalars(query).all()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError('Error loading custom roles from database!', e=e) from None


def get_custom_roles(
    session: Session, role_ids: Iterable[int] | None = None, names: Iterable[str] | None = None
) -> Sequence[CustomRole]:
    r"""Get custom roles by `role_ids` or unique `names`.

    Either the `role_ids` or `names` parameter must be specified,
    where `role_ids` will take precedence over `names`.

    Parameters
    ----------
    session : Session
        An active database session.

    role_id : Iterable[int] or None, default None
        The role_id:s of the custom roles to filter by.

    name : Iterable[str] or None, default None
        The names of the custom roles to filter by.

    Returns
    -------
    Sequence[streamlit_passwordless.db.models.CustomRole]
        The custom roles matching `role_ids` or `names`.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the custom roles from the database.
    """

    if role_ids is not None:
        where_clause = CustomRole.role_id.in_(role_ids)
    elif names is not None:
        where_clause = CustomRole.name.in_(names)
    else:
        raise exceptions.StreamlitPasswordlessError(
            'Either "role_ids" or "names" parameter must be specified!'
        )

    try:
        return session.scalars(select(CustomRole).where(where_clause)).all()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError('Error loading custom roles from database!', e=e) from None


def create_custom_role(
    session: Session, role: models.CustomRole, commit: bool = False
) -> CustomRole:
    r"""Create a new custom role in the database.

    Parameters
    ----------
    session : Session
        An active database session.

    role : streamlit_passwordless.CustomRole
        The custom role to create.

    commit : bool, default False
        True if the added custom role should be committed after being added to the session and
        False to commit later. Note that the returned `db_role` object will be in a expired state
        if committing and will be re-loaded from the database upon next access.

    Returns
    -------
    db_role : streamlit_passwordless.db.models.CustomRole
        The database model of the created custom role.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while saving the custom role to the database.

    streamlit_passwordless.DatabaseStatementError
        If there is an error with the executed SQL statement.
    """

    db_role = CustomRole(**role.model_dump())
    session.add(db_role)

    if commit:
        error_msg = (
            f'Unable to save custom role "{role.name}" to database! '
            'Check the logs for more details.'
        )
        db_commit(session=session, error_msg=error_msg)

    return db_role



r"""The entry point of the Streamlit passwordless CLI."""

# Third party
import click

# Local
from streamlit_passwordless.cli.commands.run import run
from streamlit_passwordless.metadata import __releasedate__

COMMANDS = (run,)


@click.group(
    name='stp',
    context_settings={'help_option_names': ['-h', '--help'], 'max_content_width': 1000},
)
@click.version_option(
    message=(
        f'%(prog)s, version: %(version)s, '
        f'release date: {__releasedate__}, '
        'maintainer: Anton Lydell'
    )
)
def main() -> None:
    r"""Manage the Streamlit Passwordless user database."""


for cmd in COMMANDS:
    main.add_command(cmd)

if __name__ == '__main__':
    main()



r"""The core database functionality."""

# Standard library
import logging
import sqlite3
from typing import Any, TypeAlias

# Third party
from sqlalchemy import URL as _URL
from sqlalchemy import Engine as _Engine
from sqlalchemy import create_engine, event, make_url
from sqlalchemy.exc import SQLAlchemyError, StatementError
from sqlalchemy.orm import DeclarativeBase, sessionmaker
from sqlalchemy.orm import Session as _Session

# Local
from streamlit_passwordless import exceptions
from streamlit_passwordless.database.models import Base

Engine: TypeAlias = _Engine
Session: TypeAlias = _Session
SessionFactory: TypeAlias = sessionmaker[_Session]
URL: TypeAlias = str | _URL

logger = logging.getLogger(__name__)


@event.listens_for(_Engine, "connect")
def enable_sqlite_foreign_keys(dbapi_connection, connection_record) -> None:
    r"""Enable foreign key constraints in SQLite when a new connection is established."""

    if dbapi_connection.__module__ != 'sqlite3':
        return

    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("PRAGMA foreign_keys=ON")
    except sqlite3.Error as e:
        error_msg = (
            'Error enabling foreign keys for SQLite!\n'
            'Is your version of SQLite < v3.6.19 or compiled with the SQLITE_OMIT_FOREIGN_KEY '
            'or SQLITE_OMIT_TRIGGER symbols enabled?\n'
            f'{e.__module__}.{e.__class__.__name__}({e.args[0]})'
        )
        logger.error(error_msg)

    cursor.close()


def create_session_factory(
    url: URL,
    autoflush: bool = False,
    expire_on_commit: bool = False,
    create_database: bool = True,
    base: type[DeclarativeBase] | None = None,
    connect_args: dict[Any, Any] | None = None,
    **engine_config: Any,
) -> SessionFactory:
    r"""Create the database session factory, which can produce database sessions.

    The database engine is bound to each session produced by the factory.

    Parameters
    ----------
    url : str or sqlalchemy.URL
        The SQLAlchemy database url.

    autoflush : bool, default False
        Automatically flush pending changes within the session to the database
        before executing new SQL statements.

    expire_on_commit : bool, default False
        If True make the connection between the models and the database expire after a
        transaction within a session has been committed and if False make the database models
        accessible after the commit.

    create_database : bool, default True
        If True the database table schema will be created if it does not exist.

    base : type[sqlalchemy.orm.DeclarativeBase] or None, default None
        The declarative base class that contains the metadata about the
        database table schema to create if `create_database` is True. If
        None :class:`streamlit_passwordless.db.models.Base` will be used.
        :class:`streamlit_passwordless.db.models.Base` can be subclassed
        to include additional tables in the database.

    connect_args : dict[Any, Any] or None, default None
        Additional arguments sent to the driver upon connection that further
        customizes the connection.

    **engine_config : Any
        Additional keyword arguments passed to the :func:`sqlalchemy.create_engine` function.

    Returns
    -------
    session_factory : streamlit_passwordless.db.SessionFactory
        The session factory that can produce new database sessions.
    """

    engine = create_engine(
        url=url, connect_args=connect_args if connect_args else {}, **engine_config
    )
    session_factory = sessionmaker(
        bind=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
    )

    if create_database:
        _base = Base if base is None else base
        _base.metadata.create_all(bind=engine)

    return session_factory


def commit(session: Session, error_msg: str = 'Error committing transaction!') -> None:
    r"""Commit a database transaction.

    session : streamlit_passwordless.db.Session
        An active database session.

    error_msg : str, default 'Error committing transaction!'
        An error message to add to the raised exception if an error occurs.

    Returns
    -------
    None

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while processing the database transaction.

    streamlit_passwordless.DatabaseStatementError
        If there is an error with the executed SQL statement(s).
    """

    try:
        session.commit()
    except StatementError as e:
        raise exceptions.DatabaseStatementError(message=error_msg, e=e) from None
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError(message=error_msg, e=e) from None


def create_db_url(url: str | URL) -> URL:
    r"""Create and validate the database url.

    Parameters
    ----------
    url : str or sqlalchemy.URL
        The SQLAlchemy database url. If `url` is of type
        :class:`sqlalchemy.URL` it will be returned as is.

    Returns
    -------
    sqlalchemy.URL
        The converted SQLAlchemy database url.

    Raises
    ------
    streamlit_passwordless.DatabaseInvalidUrlError
        If `url` is invalid.
    """

    try:
        return make_url(url)
    except SQLAlchemyError as e:
        raise exceptions.DatabaseInvalidUrlError(message=str(e), url=url, e=e)  # type: ignore



r"""The streamlit-passwordless package.

streamlit-passwordless provides a user model for Streamlit applications based on the Bitwarden
passwordless technology. It allows users to securely authenticate with a Streamlit application
using the passkey FIDO2 and WebAuthn protocols.
"""

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app import admin_page, init_page, load_config, setup
from streamlit_passwordless.authorization import authorized, sign_out
from streamlit_passwordless.bitwarden_passwordless import (
    BITWARDEN_PASSWORDLESS_API_URL,
    BitwardenPasswordlessClient,
    BitwardenRegisterConfig,
    PasskeyCredential,
    create_bitwarden_passwordless_client,
    register_button,
    sign_in_button,
)
from streamlit_passwordless.components import (
    ICON_ERROR,
    ICON_INFO,
    ICON_SUCCESS,
    ICON_WARNING,
    SK_CUSTOM_ROLES,
    SK_DB_USER,
    SK_REGISTER_FORM_IS_VALID,
    SK_ROLES,
    SK_SESSION_STATE_INITIALIZED,
    SK_USER,
    SK_USER_SIGN_IN,
    BannerContainer,
    BannerMessageType,
    Redirectable,
    bitwarden_register_form,
    bitwarden_register_form_existing_user,
    bitwarden_sign_in_button,
    bitwarden_sign_in_form,
    create_user_form,
    delete_user_button,
    display_banner_message,
    get_current_user,
    init_session_state,
    process_form_validation_errors,
    sign_out_button,
)
from streamlit_passwordless.config import (
    STP_BWP_PRIVATE_KEY,
    STP_BWP_PUBLIC_KEY,
    STP_BWP_URL,
    STP_DB_URL,
    STP_SECRETS_SECTION,
    ConfigManager,
)
from streamlit_passwordless.exceptions import (
    DatabaseCreateUserError,
    DatabaseError,
    DatabaseInvalidUrlError,
    DatabaseStatementError,
    RegisterUserError,
    SignInTokenVerificationError,
    StreamlitPasswordlessError,
)
from streamlit_passwordless.metadata import __releasedate__, __version__, __versiontuple__
from streamlit_passwordless.models import (
    AdminRole,
    CustomRole,
    Email,
    Role,
    SuperUserRole,
    User,
    UserRole,
    UserSignIn,
    ViewerRole,
)

# The Public API
__all__ = [
    # app
    'admin_page',
    'init_page',
    'load_config',
    'setup',
    # authorization
    'authorized',
    'sign_out',
    # bitwarden_passwordless
    'BITWARDEN_PASSWORDLESS_API_URL',
    'BitwardenPasswordlessClient',
    'BitwardenRegisterConfig',
    'create_bitwarden_passwordless_client',
    'PasskeyCredential',
    'register_button',
    'sign_in_button',
    # components
    'ICON_ERROR',
    'ICON_INFO',
    'ICON_SUCCESS',
    'ICON_WARNING',
    'SK_CUSTOM_ROLES',
    'SK_DB_USER',
    'SK_REGISTER_FORM_IS_VALID',
    'SK_ROLES',
    'SK_SESSION_STATE_INITIALIZED',
    'SK_USER',
    'SK_USER_SIGN_IN',
    'BannerContainer',
    'BannerMessageType',
    'Redirectable',
    'bitwarden_register_form',
    'bitwarden_register_form_existing_user',
    'bitwarden_sign_in_button',
    'bitwarden_sign_in_form',
    'create_user_form',
    'delete_user_button',
    'display_banner_message',
    'get_current_user',
    'init_session_state',
    'process_form_validation_errors',
    'sign_out_button',
    # config
    'STP_BWP_PRIVATE_KEY',
    'STP_BWP_PUBLIC_KEY',
    'STP_BWP_URL',
    'STP_DB_URL',
    'STP_SECRETS_SECTION',
    'ConfigManager',
    # database
    'db',
    # exceptions
    'DatabaseError',
    'DatabaseCreateUserError',
    'DatabaseInvalidUrlError',
    'DatabaseStatementError',
    'RegisterUserError',
    'SignInTokenVerificationError',
    'StreamlitPasswordlessError',
    # metadata
    '__releasedate__',
    '__version__',
    '__versiontuple__',
    # models
    'AdminRole',
    'CustomRole',
    'Email',
    'Role',
    'SuperUserRole',
    'User',
    'UserRole',
    'UserSignIn',
    'ViewerRole',
]



r"""The controller of the init page."""

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.config import ConfigManager
from streamlit_passwordless.app.views.init import create_admin_user, initialize, title
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient


def controller(
    cm: ConfigManager, db_session: db.Session, client: BitwardenPasswordlessClient
) -> None:
    r"""Render the init page.

    Parameters
    ----------
    cm : streamlit_passwordless.ConfigManager
        The loaded configuration of the application.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.
    """

    title()
    initialize(db_session=db_session, db_url=cm.db_url)
    create_admin_user(db_session=db_session, client=client)



r"""Database operations on the user_sign_in table."""

# Standard library

# Third party

# Local
from .. import models
from ..core import Session
from ..core import commit as db_commit
from ..schemas import user_sign_in as schemas


def create_user_sign_in(
    session: Session, user_sign_in: schemas.UserSignInCreate, commit: bool = True
) -> models.UserSignIn:
    r"""Create a new user sign in entry in the database.

    Parameters
    ----------
    session : streamlit_passwordless.db.Session
        An active database session.

    user_sign_in : streamlit_passwordless.db.UserSignInCreate
        The user sign in entry to create.

    commit : bool, default True
        True if the added user sign in entry should be committed after being added to the
        session and False to commit later. Note that the returned `db_user_sign_in` object
        will be in a expired state if committing and will be re-loaded from the database upon
        next access.

    Returns
    -------
    db_user_sign_in : streamlit_passwordless.db.models.UserSignIn
        The created user sign in entry.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while saving the user sign in entry to the database.

    streamlit_passwordless.DatabaseStatementError
        If there is an error with the executed SQL statement.
    """

    db_user_sign_in = models.UserSignIn(**user_sign_in.model_dump())
    session.add(db_user_sign_in)

    if commit:
        error_msg = 'Unable to save user sign in to database! Check the logs for more details.'
        db_commit(session=session, error_msg=error_msg)

    return db_user_sign_in



r"""The ID:s of components used for identification and session state management."""

# =====================================================================================
# Create user form
# =====================================================================================

CREATE_USER_FORM = 'create-user-form'
CREATE_USER_FORM_USERNAME_TEXT_INPUT = 'create-user-form-username-text-input'
CREATE_USER_FORM_DISPLAYNAME_TEXT_INPUT = 'create-user-form-displayname-text-input'
CREATE_USER_FORM_AD_USERNAME_TEXT_INPUT = 'create-user-form-ad-username-text-input'
CREATE_USER_FORM_ROLE_SELECTBOX = 'create-user-form-role-selectbox'
CREATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX = 'create-user-form-custom-roles-multiselectbox'
CREATE_USER_FORM_EMAIL_TEXT_INPUT = 'create-user-form-email-text-input'
CREATE_USER_FORM_DISABLED_TOGGLE = 'create-user-form-disabled-toggle'
CREATE_USER_FORM_SUBMIT_BUTTON = 'create-user-form-submit-button'

# =====================================================================================
# Delete user
# =====================================================================================

DELETE_USER_BUTTON = 'delete-user-button'

# =====================================================================================
# Register form
# =====================================================================================

BP_REGISTER_FORM = 'bp-register-form'
BP_REGISTER_FORM_USERNAME_TEXT_INPUT = 'bp-register-form-username-text-input'
BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT = 'bp-register-form-displayname-text-input'
BP_REGISTER_FORM_EMAIL_TEXT_INPUT = 'bp-register-form-email-text-input'
BP_REGISTER_FORM_CREDENTIAL_NICKNAME_TEXT_INPUT = 'bp-register-form-credential-nickname-text-input'
BP_REGISTER_FORM_DISCOVERABILITY_RADIO_BUTTON = 'bp-register-form-discoverability-radio-button'
BP_REGISTER_FORM_DISCOVERABILITY_TOGGLE_SWITCH = 'bp-register-form-discoverability-toggle-switch'
BP_REGISTER_FORM_ALIASES_TEXT_INPUT = 'bp-register-form-aliases-text-input'
BP_REGISTER_FORM_SUBMIT_BUTTON = 'bp-register-form-submit-button'
BP_REGISTER_FORM_EXISTING_USER_CREDENTIAL_NICKNAME_TEXT_INPUT = (
    'bp-register-form-existing-user-credential-nickname-text-input'
)
BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_RADIO_BUTTON = (
    'bp-register-form-existing-user-discoverability-radio-button'
)
BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_TOGGLE_SWITCH = (
    'bp-register-form-existing-user-discoverability-toggle-switch'
)
BP_REGISTER_FORM_EXISTING_USER_SUBMIT_BUTTON = 'bp-register-form-existing-user-submit-button'

# =====================================================================================
# Sign in form
# =====================================================================================

BP_SIGN_IN_BUTTON = 'bp-sign-in-button'
BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT = 'bp-sign-in-form-alias-text-input'
BP_SIGN_IN_FORM_SUBMIT_BUTTON = 'bp-sign-in-form-submit-button'

# =====================================================================================
# Sign out
# =====================================================================================

KEY_SIGN_OUT_BUTTON = 'stp-sign-out-button'

# =====================================================================================
# Update user form
# =====================================================================================

UPDATE_USER_FORM = 'update-user-form'
UPDATE_USER_FORM_USERNAME_TEXT_INPUT = 'update-user-form-username-text-input'
UPDATE_USER_FORM_DISPLAYNAME_TEXT_INPUT = 'update-user-form-displayname-text-input'
UPDATE_USER_FORM_AD_USERNAME_TEXT_INPUT = 'update-user-form-ad-username-text-input'
UPDATE_USER_FORM_ROLE_SELECTBOX = 'update-user-form-role-selectbox'
UPDATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX = 'update-user-form-custom-roles-multiselectbox'
UPDATE_USER_FORM_DISABLED_TOGGLE = 'update-user-form-disabled-toggle'
UPDATE_USER_FORM_SUBMIT_BUTTON = 'update-user-form-submit-button'



r"""The configuration of the Streamlit Passwordless web app."""

# Standard library
from enum import StrEnum

# Local
from streamlit_passwordless.bitwarden_passwordless import (
    BitwardenPasswordlessClient,
    create_bitwarden_passwordless_client,
)
from streamlit_passwordless.components.config import init_session_state
from streamlit_passwordless.config import ConfigManager, load_config
from streamlit_passwordless.database import (
    SessionFactory,
    create_session_factory_cached,
)
from streamlit_passwordless.metadata import __releasedate__, __version__

MAINTAINER_INFO = f"""\
- Maintainer   : [*Anton Lydell*](https://github.com/antonlydell)
- Version      : {__version__}
- Release date : {__releasedate__}"""

APP_HOME_PAGE_URL = 'https://github.com/antonlydell/streamlit-passwordless'
APP_ISSUES_PAGE_URL = 'https://github.com/antonlydell/streamlit-passwordless/issues'


class Pages(StrEnum):
    r"""The pages of the application."""

    ADMIN = 'pages/admin.py'
    INIT = 'pages/init.py'
    SIGN_IN = 'pages/sign_in.py'


def setup(
    create_database: bool = False,
) -> tuple[BitwardenPasswordlessClient, SessionFactory, ConfigManager]:
    r"""Setup the resources needed to run a Streamlit Passwordless application.

    Parameters
    ----------
    create_database : bool, default False
        If True the database tables are created and if False table creation is omitted.

    Returns
    -------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    session_factory : streamlit_passwordless.db.SessionFactory
        The session factory that can produce new database sessions.

    cm : streamlit_passwordless.ConfigManager
        The loaded configuration of the application.
    """

    init_session_state()
    cm = load_config()
    client = create_bitwarden_passwordless_client(
        public_key=cm.bwp_public_key, private_key=cm.bwp_private_key, _url=cm.bwp_url
    )
    session_factory = create_session_factory_cached(url=cm.db_url, create_database=create_database)

    return client, session_factory, cm



r"""The create user form component and its callback functions."""

# Standard library
import logging
from typing import Sequence

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless import models
from streamlit_passwordless.common import get_current_datetime

from . import config, core, ids

logger = logging.getLogger(__name__)


def _validate_form(db_session: db.Session, email_is_username: bool) -> None:
    r"""Validate the input fields of the create user form.

    Updates the following session state keys:
    - `streamlit_passwordless.SK_CREATE_USER_FORM_IS_VALID` :
        True if the form validation passed and False otherwise.

    - `streamlit_passwordless.SK_CREATE_USER_FORM_VALIDATION_ERRORS` :
        A dictionary mapping of the user form field name to its error message.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    email_is_username : bool
        If True the username is the email address of the user and if False
        the username is distinct from an optional email address of the user.
    """

    form_is_valid = True
    validation_errors = {}

    validation_error_field = core.FormField.USERNAME
    if username := st.session_state[ids.CREATE_USER_FORM_USERNAME_TEXT_INPUT]:
        db_user, error_msg = core.get_user_from_database(
            session=db_session, username=username.strip().casefold()
        )

        if error_msg:  # DatabaseError
            form_is_valid = False
            validation_errors[validation_error_field] = error_msg
        elif db_user is not None:
            form_is_valid = False
            validation_errors[validation_error_field] = f'User {username} already exists!'
        else:
            pass

    else:
        validation_errors[validation_error_field] = 'The username field is required!'

    if email_is_username:
        email = username
        validation_error_field = core.FormField.USERNAME
    else:
        email = st.session_state.get(ids.CREATE_USER_FORM_EMAIL_TEXT_INPUT)
        validation_error_field = core.FormField.EMAIL

    if email:
        db_email, error_msg = core.get_email_from_database(
            session=db_session, email=email.strip().casefold()
        )
        if error_msg:  # DatabaseError
            form_is_valid = False
            validation_errors[validation_error_field] = error_msg
        elif db_email is not None:
            form_is_valid = False
            validation_errors[validation_error_field] = (
                f'{"User" if email_is_username else "Email"} {email} already exists!'
            )
        else:
            pass

    st.session_state[config.SK_CREATE_USER_FORM_VALIDATION_ERRORS] = validation_errors
    st.session_state[config.SK_CREATE_USER_FORM_IS_VALID] = form_is_valid


def create_user_form(
    db_session: db.Session,
    with_displayname: bool = True,
    with_ad_username: bool = False,
    with_role: bool = True,
    with_custom_roles: bool = True,
    with_email: bool = True,
    with_disabled_toggle: bool = True,
    email_is_username: bool = False,
    roles: Sequence[models.Role] | None = None,
    custom_roles: Sequence[db.models.CustomRole] | None = None,
    title: str = '#### Create a new user',
    border: bool = True,
    submit_button_label: str = 'Create User',
    submit_button_type: core.ButtonType = 'primary',
    clear_on_submit: bool = False,
    banner_container: core.BannerContainer | None = None,
    created_by_user_id: str | None = None,
    username_label: str = 'Username',
    username_max_length: int | None = 50,
    username_placeholder: str | None = 'john.doe',
    username_help: str | None = 'A unique identifier for the user.',
    displayname_label: str = 'Displayname',
    displayname_max_length: int | None = 50,
    displayname_placeholder: str | None = 'John Doe',
    displayname_help: str | None = 'A descriptive name of the user.',
    ad_username_label: str = 'Active Directory Username',
    ad_username_max_length: int | None = 50,
    ad_username_placeholder: str | None = 'ad.john.doe',
    ad_username_help: str | None = 'The active directory username of the user.',
    role_label: str = 'Role',
    role_placeholder: str = 'Choose a role',
    role_preselected: int | None = 0,
    role_help: str | None = 'The role of the user.',
    custom_roles_label: str = 'Custom Roles',
    custom_roles_placeholder: str = 'Choose a custom role',
    custom_roles_max_selections: int | None = None,
    custom_roles_default_selection: (
        Sequence[db.models.CustomRole] | db.models.CustomRole | None
    ) = None,
    custom_roles_help: str | None = 'The custom roles to associate with the user.',
    email_label: str = 'Email',
    email_max_length: int | None = 50,
    email_placeholder: str | None = 'john.doe@example.com',
    email_help: str | None = 'An email address to associate with the user.',
    disabled_toggle_label: str = 'Disabled User',
    disabled_toggle_default_value: bool = False,
    disabled_toggle_help: str | None = 'If the user should be created as disabled or enabled.',
) -> models.User | None:
    r"""Render the create user form.

    Create a new user in the database without registering a passkey.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    with_displayname : bool, default True
        If True the displayname field is added to the form to enable to
        enter a displayname for the user.

    with_ad_username : bool, default False
        If True the ad_username field is added to the form to enable to
        enter the active directory username for the user.

    with_role : bool, default True
        If True the role selectbox is added to the form to enable to set
        the desired role for the new user.

    with_custom_roles : bool, default False
        If True the custom roles multiselectbox is added to the form, given that the
        `custom_roles` parameter is specified, to enable to set the desired custom roles
        for the new user.

    with_email : bool, default True
        If True the email field is added to the form to enable to
        enter an email address to associate with the user.

    with_disabled_toggle : bool, default True
        If True the disabled user toggle switch will be added to the form, which enables
        creating a new user as disabled or enabled. If False a new user will be created as enabled.

    email_is_username : bool, default False
        If True the username field will prompt to enter an email address, which will be used
        as the username and be added as an email address of the user.

    roles : Sequence[streamlit_passwordless.Role] or None, default None
        The available roles that can be assigned to the user. If None the
        default roles of Streamlit Passwordless will be used.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole] or None, default None
        The available custom roles that can be assigned to the user. If None the defined
        custom roles in the database will be loaded and made available for selection.

    title : str, default '#### Create a new user'
        The title of the create user form. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    submit_button_label : str, default 'Create User'
        The label of the form submit button.

    submit_button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the submit button. Emulates the `type` parameter of :func:`streamlit.button`.

    clear_on_submit : bool, default False
        If True the form fields will be cleared when the submit form button is clicked.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the create user process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the form.

    created_by_user_id : str or None, default None
        The ID of the user that is creating the new user.
        If None the ID of the currently signed in user is used.

    Other Parameters
    ----------------
    username_label : str, default 'Username'
        The label of the username field.

    username_max_length : int or None, default 50
        The maximum allowed number of characters of the username field.
        If None the upper limit is removed.

    username_placeholder : str or None, default 'john.doe'
        The placeholder of the username field. If None the placeholder is removed.

    username_help : str or None, default 'A unique identifier for the user.'
        The help text to display for the username field. If None the help text is removed.

    displayname_label : str, default 'Displayname'
        The label of the displayname field.

    displayname_max_length : int or None, default 50
        The maximum allowed number of characters of the displayname field.
        If None the upper limit is removed.

    displayname_placeholder : str or None, default 'John Doe'
        The placeholder of the displayname field. If None the placeholder is removed.

    displayname_help : str or None, default 'A descriptive name of the user.'
        The help text to display for the displayname field. If None the help text is removed.

    ad_username_label : str, default 'Active Directory Username'
        The label of the ad_username field.

    ad_username_max_length : int or None, default 50
        The maximum allowed number of characters of the ad_username field.
        If None the upper limit is removed.

    ad_username_placeholder : str or None, default 'ad.john.doe'
        The placeholder of the ad_username field. If None the placeholder is removed.

    ad_username_help : str or None, default 'The active directory username of the user.'
        The help text to display for the ad_username field. If None the help text is removed.

    role_label : str, default 'Role'
        The label of the role selectbox.

    role_placeholder : str, default 'Choose a role'
        The placeholder of the role selectbox if a role is not preselected.

    role_preselected : int or None, default 0
        The index value of `roles` that will be used to extract the default selected role of
        the role selectbox. If None no role will be selected by default on first render.

    role_help: str or None, default 'The role of the user.'
        The help text to display for the role selectbox. If None the help text is removed.

    custom_roles_label : str, default 'Custom Roles',
        The label of the custom roles multiselectbox.

    custom_roles_placeholder : str, default 'Choose a custom role'
        The placeholder of the custom roles multiselectbox if no custom roles are preselected.
        If None the placeholder is removed.

    custom_roles_max_selections : int or None, default None
        The maximum number of custom roles that can be selected in the custom roles multiselectbox.
        If None there is no upper limit.

    custom_roles_default_selection : Sequence[streamlit_passwordless.db.models.CustomRole] or streamlit_passwordless.models.db.CustomRole or None, default None
        The custom roles that will be preselected in the the custom roles mulitselectbox.
        If None no custom roles will be preselected on first render.

    custom_roles_help : str or None, default 'The custom roles to associate with the user.'
        The help text to display for the custom roles multiselectbox.
        If None the help text is removed.

    email_label : str, default 'Email'
        The label of the email field.

    email_max_length : int or None, default 50
        The maximum allowed number of characters of the email field.
        If None the upper limit is removed.

    email_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the email field. If None the placeholder is removed.

    email_help : str or None, default 'An email address to associate with the user.'
        The help text to display for the email field. If None the help text is removed.

    disabled_toggle_label : str, default 'Disabled user'
        The label for the disabled user toggle switch.

    disabled_toggle_default_value : bool, default False
        The default value of the disabled user toggle switch. True means a user will be created
        as disabled and False as enabled.

    disabled_toggle_help : str or None, default 'If the user should be created as disabled or enabled.'
        The help text to display for the disabled toggle switch. If None the help text is removed.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The created user. None is returned if a user could not be created.
    """

    banner_container = st.empty() if banner_container is None else banner_container
    banner_container_mapping = {}

    with st.form(key=ids.CREATE_USER_FORM, clear_on_submit=clear_on_submit, border=border):
        st.markdown(title)

        if email_is_username:
            username_label = email_label
            username_placeholder = email_placeholder
            username_max_length = email_max_length

        username_error_banner = st.empty()
        banner_container_mapping[core.FormField.USERNAME] = username_error_banner

        username = st.text_input(
            label=username_label,
            placeholder=username_placeholder,
            max_chars=username_max_length,
            help=username_help,
            key=ids.CREATE_USER_FORM_USERNAME_TEXT_INPUT,
        )

        if with_displayname:
            displayname = st.text_input(
                label=displayname_label,
                placeholder=displayname_placeholder,
                max_chars=displayname_max_length,
                help=displayname_help,
                key=ids.CREATE_USER_FORM_DISPLAYNAME_TEXT_INPUT,
            )
        else:
            displayname = None

        if with_ad_username:
            ad_username = st.text_input(
                label=ad_username_label,
                placeholder=ad_username_placeholder,
                max_chars=ad_username_max_length,
                help=ad_username_help,
                key=ids.CREATE_USER_FORM_AD_USERNAME_TEXT_INPUT,
            )
        else:
            ad_username = None

        if with_role:
            if roles is None:
                roles = (models.ViewerRole, models.UserRole, models.SuperUserRole, models.AdminRole)
            role = st.selectbox(
                label=role_label,
                options=roles,
                index=role_preselected,
                placeholder=role_placeholder,
                format_func=lambda r: r.name,
                help=role_help,
                key=ids.CREATE_USER_FORM_ROLE_SELECTBOX,
            )
        else:
            role = models.UserRole

        if with_custom_roles:
            if custom_roles is None:
                db_custom_roles, error_msg = core.get_all_custom_roles_from_database(
                    session=db_session
                )
                if error_msg:
                    core.display_banner_message(
                        message=error_msg,
                        message_type=core.BannerMessageType.ERROR,
                        container=st.empty(),
                    )
            else:
                db_custom_roles = custom_roles
            selected_custom_roles: Sequence[db.models.CustomRole] = st.multiselect(
                label=custom_roles_label,
                options=db_custom_roles,
                default=custom_roles_default_selection,
                max_selections=custom_roles_max_selections,
                placeholder=custom_roles_placeholder,
                format_func=lambda r: r.name,
                help=custom_roles_help,
                key=ids.CREATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX,
            )
        else:
            selected_custom_roles = tuple()
        if with_email:
            if not email_is_username:
                email_error_banner = st.empty()
                banner_container_mapping[core.FormField.EMAIL] = email_error_banner
                email = st.text_input(
                    label=email_label,
                    placeholder=email_placeholder,
                    max_chars=email_max_length,
                    help=email_help,
                    key=ids.CREATE_USER_FORM_EMAIL_TEXT_INPUT,
                )
            else:
                email = username
        else:
            email = None

        if with_disabled_toggle:
            disabled = st.toggle(
                label=disabled_toggle_label,
                value=disabled_toggle_default_value,
                help=disabled_toggle_help,
                key=ids.CREATE_USER_FORM_DISABLED_TOGGLE,
            )
        else:
            disabled = False

        clicked = st.form_submit_button(
            label=submit_button_label,
            type=submit_button_type,
            on_click=_validate_form,
            kwargs={'db_session': db_session, 'email_is_username': email_is_username},
        )

        if not clicked:
            return None

        if not st.session_state.get(config.SK_CREATE_USER_FORM_IS_VALID):
            validation_error_key = config.SK_CREATE_USER_FORM_VALIDATION_ERRORS
            core.process_form_validation_errors(
                validation_errors=st.session_state[validation_error_key],
                banner_container_mapping=banner_container_mapping,  # type: ignore
                default_banner_container=banner_container,
            )
            st.session_state[validation_error_key] = {}

            return None

    user = models.User(
        username=username.strip(),
        ad_username=None if ad_username is None else ad_username.strip(),
        displayname=None if displayname is None else displayname.strip(),
        disabled=disabled,
        disabled_at=get_current_datetime() if disabled else None,
        role=models.UserRole if role is None else role,
        emails=[models.Email(email=email.strip().casefold(), rank=1)] if email else [],
        custom_roles={m.name: models.CustomRole.model_validate(m) for m in selected_custom_roles},
    )

    if created_by_user_id is None:
        signed_in_user = config.get_current_user()
        created_by = signed_in_user.user_id if signed_in_user else None
    else:
        created_by = created_by_user_id

    success, error_msg = core.create_user_in_database(
        session=db_session, user=user, created_by_user_id=created_by
    )
    if not success:
        core.display_banner_message(
            message=error_msg, message_type=core.BannerMessageType.ERROR, container=banner_container
        )
        return None

    msg = f'Successfully created user: {user.username}!'
    logger.info(msg)
    core.display_banner_message(
        message=msg, message_type=core.BannerMessageType.SUCCESS, container=banner_container
    )

    return user



r"""The sign in form component."""

# Standard library
import logging
from typing import Any

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.bitwarden_passwordless.backend import BitwardenPasswordlessClient
from streamlit_passwordless.bitwarden_passwordless.frontend import sign_in_button

from . import config, core, ids

logger = logging.getLogger(__name__)


def _get_user_from_db(
    session: db.Session, user_sign_in: models.UserSignIn
) -> tuple[models.User | None, str]:
    r"""Get the signed in user from the database.

    The following session state keys are updated with data about the user that signed in:
    - config.SK_DB_USER : The database user object.
    - config.SK_USER : The user object.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    user_sign_in : streamlit_passwordless.UserSignIn
        Data from Bitwarden Passwordless about the user that signed in.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The user that signed in. None is returned if no user with a
        matching user_id from `user_sign_in` was found in the database.

    error_msg : str
        An error message with info if the user could not be loaded from the database.
        If no error occurred an empty string is returned.
    """

    db_user = None
    user = None
    try:
        db_user = db.get_user_by_user_id(session=session, user_id=user_sign_in.user_id)
        load_db_user_error = False
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        load_db_user_error = True
    else:
        error_msg = ''

    if db_user is None:
        if not load_db_user_error:
            logger.error(
                f'Signed in user (user_id={user_sign_in.user_id}, '
                f'credential_nickname={user_sign_in.credential_nickname}) '
                'was not found in local database!\n'
                'A mismatch between Bitwarden Passwordless.dev and local database has occurred!'
            )
            error_msg = (
                f'User trying to sign in with passkey {user_sign_in.credential_nickname} '
                'does not exist! Check the logs for more details.'
            )
    else:
        user = models.User.model_validate(db_user)

    st.session_state[config.SK_DB_USER] = db_user
    st.session_state[config.SK_USER] = user

    return user, error_msg


def _authorize_user(user: models.User, role: models.Role | int | None) -> tuple[models.User, str]:
    r"""Check if the `user` is authorized to sign in to the application.

    Parameters
    ----------
    user : streamlit_passwordless.User
        The user to authorize.

    role : streamlit_passwordless.Role or int or None
        The role to authorize `user` against. If the rank of the role of `user` is
        greater than or equal to the rank of `role` the `user` is authorized. If an
        integer is supplied it is assumed to be the rank of the role to authorize the
        user against. If None (the default) the user is authorized regardless of its role.
        The user must always be authenticated to be authorized.

    Returns
    -------
    user : streamlit_passwordless.User
        The authorized or not authorized `user`, where the `sign_in`
        attribute has been set to None if the user is not authorized.

    message : str
        An error message if the user was not authorized.
        An empty string is returned if the user was authorized.
    """

    if user.is_authorized(role=role):
        return user, ''

    user.sign_in = None

    if role is None:
        role_info = 'None'
    else:
        role_name = role.name if isinstance(role, models.Role) else ''
        role_rank = role.rank if isinstance(role, models.Role) else role
        role_info = f'Role(name={role_name}, rank={role_rank})'

    warning_msg = (
        f'User(username={user.username}, role=Role(name={user.role.name}, '
        f'rank={user.role.rank}) is not authorized to sign in '
        f'against role {role_info}'
    )
    logger.warning(warning_msg)

    return user, f'User {user.username} is not authorized to sign in!'


def _process_user_sign_in(
    token: str,
    error: dict[str, Any] | None,
    clicked: bool,
    client: BitwardenPasswordlessClient,
    session: db.Session,
    role: models.Role | int | None,
    banner_container: core.BannerContainer,
    redirect: core.Redirectable | None,
) -> tuple[models.User | None, bool]:
    r"""Process the user sign in and authorize the user.

    Parameters
    ----------
    token : str
        The verification token to be used by the Bitwarden Passwordless
        backend API to authenticate the sign in process.

    error : dict[str, Any] or None
        An error object from the frontend containing information if there
        was an error with the sign in process.

    clicked : bool
        True if the sign in button was clicked and False otherwise.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless backend API.

    session : streamlit_passwordless.db.Session
        An active database session.

    role : streamlit_passwordless.Role or int or None
        The role to authorize `user` against. If the rank of the role of `user` is
        greater than or equal to the rank of `role` the `user` is authorized. If an
        integer is supplied it is assumed to be the rank of the role to authorize the
        user against. If None (the default) the user is authorized regardless of its role.
        The user must always be authenticated to be authorized.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the sign in process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the button or form.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to if authorized. If str it should be
        the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None no redirect is performed.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The user that signed in and was authorized or not authorized. None is returned
        if no user with a matching user_id from `user_sign_in` was found in the database.

    success : bool
        True if the user was signed in and authorized without errors and False otherwise.
    """

    if not clicked:
        return None, False

    if not token and error:
        error_msg = f'Error signing in!\nerror : {error}'
        logger.error(error_msg)
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )
        return None, False

    user_sign_in, error_msg = core.verify_sign_in(client=client, token=token)

    if user_sign_in is None or user_sign_in.success is False:
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )
        return None, False

    user, error_msg = _get_user_from_db(session=session, user_sign_in=user_sign_in)
    if user is None:
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )
        return user, False
    else:
        user.sign_in = user_sign_in

    user, error_msg = _authorize_user(user=user, role=role)
    if error_msg:
        user.sign_in = None
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )
        return user, False

    success, error_msg = core.save_user_sign_in_to_database(
        session=session, user_sign_in=user_sign_in
    )

    if not success:
        user.sign_in = None
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )
        return user, success

    if redirect:
        st.switch_page(redirect)

    core.display_banner_message(
        message=f'Successfully signed in user {user.username}',
        message_type=core.BannerMessageType.SUCCESS,
        container=banner_container,
    )

    return user, success


def bitwarden_sign_in_form(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    role: models.Role | int | None = None,
    with_alias: bool = True,
    with_discoverable: bool = True,
    with_autofill: bool = False,
    title: str = '#### Sign in',
    border: bool = True,
    submit_button_label: str = 'Sign in',
    button_type: core.ButtonType = 'primary',
    banner_container: core.BannerContainer | None = None,
    redirect: core.Redirectable | None = None,
    alias_label: str = 'Alias',
    alias_max_length: int | None = 50,
    alias_placeholder: str | None = 'john.doe@example.com',
    alias_help: str | None = '__default__',
) -> tuple[models.User | None, bool]:
    r"""Render the Bitwarden Passwordless sign in form.

    Allows the user to sign in to the application with a registered passkey.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    db_session : streamlit_passwordless.db.Session
        An active database session.

    role : streamlit_passwordless.Role or int or None, default None
        The role to authorize `user` against. If the rank of the role of `user` is
        greater than or equal to the rank of `role` the `user` is authorized. If an
        integer is supplied it is assumed to be the rank of the role to authorize the
        user against. If None (the default) the user is authorized regardless of its role.
        The user must always be authenticated to be authorized.

    with_alias : bool, default True
        If True the field to enter the alias to use for signing in will be rendered. If False
        the field is not rendered and the other sign in methods `with_discoverable` or
        `with_autofill` should be used. If the user specifies an alias it will override the other
        sign in methods `with_discoverable` and `with_autofill`.

    with_discoverable : bool, default True
        If True the browser's native UI prompt will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. If `alias` is specified it will
        override this sign in method.

    with_autofill : bool, default False
        If True the browser's native autofill UI will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. This method of signing in is
        overridden if `alias` specified or `with_discoverable` is True.

    title : str, default '#### Sign in.'
        The title of the sign in from. Markdown is supported.

    border : bool, default True
        True if a border surrounding the form should be rendered and False
        to remove the border.

    submit_button_label : str, default 'Sign in'
        The label of the submit button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the sign in process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the form.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to if authorized. If str it should be
        the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None no redirect is performed.

    Other Parameters
    ----------------
    alias_label : str, default 'Alias'
        The label of the alias field.

    alias_max_length : int or None, default 50
        The maximum allowed number of characters of the alias field.
        If None the upper limit is removed.

    alias_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the alias field. If None the placeholder is removed.

    alias_help : str or None, default '__default__'
        The help text to display for the alias field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The user object of the user that signed in. None is returned if a user has not
        signed in yet or if the sign in failed and a user object could not be retrieved.

    success : bool
        True if the user was signed in and authorized without errors and False otherwise.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If no sign in method is chosen i.e. `with_alias` is False,
        `with_discoverable` is False and `with_autofill` is False.
    """

    if with_alias is False and with_discoverable is False and with_autofill is False:
        error_msg = (
            f'At least one sign in method must be chosen!\n'
            f'bitwarden_sign_in_form({with_alias=}, {with_discoverable=}, {with_autofill=})'
        )
        logger.error(error_msg)
        raise exceptions.StreamlitPasswordlessError(
            message=error_msg,
            data={
                'with_alias': with_alias,
                'with_discoverable': with_discoverable,
                'with_autofill': with_autofill,
            },
        )

    help: str | None = None
    banner_container = st.empty() if banner_container is None else banner_container

    with st.container(border=border):
        st.markdown(title)

        if with_alias:
            if alias_help == '__default__':
                help = (
                    'An alias of the user. If not supplied auto discover '
                    'of available credentials will be attempted.'
                )
            else:
                help = alias_help

            st.text_input(
                label=alias_label,
                placeholder=alias_placeholder,
                max_chars=alias_max_length,
                help=help,
                key=ids.BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT,
            )

        token, error, clicked = sign_in_button(
            public_key=client.public_key,
            alias=st.session_state.get(ids.BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT),
            with_discoverable=with_discoverable,
            with_autofill=with_autofill,
            label=submit_button_label,
            button_type=button_type,
            key=ids.BP_SIGN_IN_FORM_SUBMIT_BUTTON,
        )

    user, success = _process_user_sign_in(
        token=token,
        error=error,
        clicked=clicked,
        client=client,
        session=db_session,
        role=role,
        banner_container=banner_container,
        redirect=redirect,
    )

    return user, success


def bitwarden_sign_in_button(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    role: models.Role | int | None = None,
    with_discoverable: bool = True,
    with_autofill: bool = False,
    button_label: str = 'Sign in',
    button_type: core.ButtonType = 'primary',
    banner_container: core.BannerContainer | None = None,
    redirect: core.Redirectable | None = None,
) -> tuple[models.User | None, bool]:
    r"""Render the Bitwarden Passwordless sign in button.

    Allows the user to sign in to the application with a discoverable passkey.
    Similar to :func:`bitwarden_sign_in_form`, but without the possibility to
    sign in with user_id, alias or a non-discoverable passkey.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    db_session : streamlit_passwordless.db.Session
        An active database session.

    role : streamlit_passwordless.Role or int or None, default None
        The role to authorize `user` against. If the rank of the role of `user` is
        greater than or equal to the rank of `role` the `user` is authorized. If an
        integer is supplied it is assumed to be the rank of the role to authorize the
        user against. If None (the default) the user is authorized regardless of its role.
        The user must always be authenticated to be authorized.

    with_discoverable : bool, default True
        If True the browser's native UI prompt will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. If `alias` is specified it will
        override this sign in method.

    with_autofill : bool, default False
        If True the browser's native autofill UI will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. This method of signing in is
        overridden if `alias` specified or `with_discoverable` is True.

    button_label : str, default 'Sign in'
        The label of the button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the sign in process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the button.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to if authorized. If str it should be
        the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None no redirect is performed.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The user that signed in. None is returned if a user has not signed
        in yet or if the sign in failed and a user could not be retrieved.

    success : bool
        True if the user was signed in and authorized without errors and False otherwise.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If no sign in method is chosen i.e. `with_discoverable` is False
        and `with_autofill` is False.
    """

    if with_discoverable is False and with_autofill is False:
        error_msg = (
            f'At least one sign in method must be chosen!\n'
            f'bitwarden_sign_in_button({with_discoverable=}, {with_autofill=})'
        )
        logger.error(error_msg)
        raise exceptions.StreamlitPasswordlessError(
            message=error_msg,
            data={'with_discoverable': with_discoverable, 'with_autofill': with_autofill},
        )

    banner_container = st.empty() if banner_container is None else banner_container

    token, error, clicked = sign_in_button(
        public_key=client.public_key,
        with_discoverable=with_discoverable,
        with_autofill=with_autofill,
        label=button_label,
        button_type=button_type,
        key=ids.BP_SIGN_IN_BUTTON,
    )

    user, success = _process_user_sign_in(
        token=token,
        error=error,
        clicked=clicked,
        client=client,
        session=db_session,
        role=role,
        banner_container=banner_container,
        redirect=redirect,
    )

    return user, success



r"""The controller of the sign in page."""

# Third party
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.config import Pages
from streamlit_passwordless.app.views.sign_in import title
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient
from streamlit_passwordless.components import bitwarden_sign_in_button
from streamlit_passwordless.models import AdminRole


def controller(client: BitwardenPasswordlessClient, db_session: db.Session) -> None:
    r"""Render the sign in page.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.
    """

    banner_container = st.empty()

    with st.container(border=True):
        title()
        bitwarden_sign_in_button(
            client=client,
            db_session=db_session,
            role=AdminRole,
            banner_container=banner_container,
            redirect=Pages.ADMIN,
        )



r"""The models that represent tables in the database."""

# Standard library
import os
from datetime import datetime
from enum import StrEnum
from typing import ClassVar, Optional, Self

# Third party
from sqlalchemy import TIMESTAMP, Column, ForeignKey, Index, MetaData, Table, UniqueConstraint, func
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    attribute_keyed_dict,
    mapped_column,
    relationship,
)

SCHEMA: str | None = os.getenv('STP_DB_SCHEMA')
metadata_obj = MetaData(schema=SCHEMA)

updated_at_column: Mapped[Optional[datetime]] = mapped_column(
    TIMESTAMP(), onupdate=func.current_timestamp()
)
created_at_column: Mapped[datetime] = mapped_column(
    TIMESTAMP(), server_default=func.current_timestamp()
)


def _column_to_str(col: object) -> str:
    r"""Convert a column value to string format.

    To be used for the `__repr__` methods of the models.
    """

    return col.isoformat() if isinstance(col, datetime) else repr(col)


class Base(DeclarativeBase):
    r"""The base model all database models will inherit from.

    It defines the schema/user where the tables will be created in the
    database and the columns, which are common for all tables.

    Class variables
    ---------------
    columns__repr__ : ClassVar[tuple[str, ...]], default tuple()
        The names of the columns that should be part of the `__repr__` method output.
        Each model should define this variable. If columns are defined on :class:`Base`
        they should not be part of `columns__repr__`.

    indent_space__repr__ : ClassVar[str], default ' ' * 4
        The indentation space used for each column in the `__repr__` method.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = tuple()
    indent_space__repr__: ClassVar[str] = ' ' * 4

    metadata = metadata_obj

    def __repr__(self) -> str:
        r"""A string representation of the model."""

        indent = self.indent_space__repr__
        output = f'{self.__class__.__name__}(\n'

        for col in self.columns__repr__:
            value = getattr(self, col)
            value = _column_to_str(value)
            output += f'{indent}{col}={value},\n'

        output += ')'

        return output


class ModifiedAndCreatedColumnMixin:
    r"""Add columns for when a record was last updated and created in a table.

    Parameters
    ----------
    updated_at : datetime or None
        The timestamp at which the record was last updated (UTC).

    updated_by : str or None
        The ID of the user that last updated the record.

    created_at : datetime
        The timestamp at which the record was created (UTC).
        Defaults to current timestamp.

    created_by : str or None
        The ID of the user that created the record.

    A table model should inherit from this class prior to :class:`Base`.
    E.g. creating the user model:

    .. code-block:: python

        class User(ModifiedAndCreatedColumnMixin, Base):
            pass
    """

    updated_at: Mapped[Optional[datetime]] = updated_at_column
    updated_by: Mapped[Optional[str]]
    created_at: Mapped[datetime] = created_at_column
    created_by: Mapped[Optional[str]]


class UserRoleName(StrEnum):
    r"""The predefined user role names of streamlit-passwordless.

    These roles are created in the database when the database is initialized.
    The default role of a new user is :attr:`UserRoleName.USER`.

    Members
    -------
    VIEWER
        A user that can only view data within an application.

    USER
        The standard user with normal privileges. When a user is created it is
        assigned this role by default.

    SUPERUSER
        A user with higher privileges that can perform certain
        operations that a normal `USER` can not.

    ADMIN
        An admin has full access to everything. Only admin users may sign in to the admin page
        and manage the users of the application. An application should have at least one admin.
    """

    VIEWER = 'Viewer'
    USER = 'User'
    SUPERUSER = 'SuperUser'
    ADMIN = 'Admin'


class Role(ModifiedAndCreatedColumnMixin, Base):
    r"""The role of a user.

    A :class:`User` is associated with a role to manage its privileges within an application.
    The available roles are predefined by streamlit-passwordless, but can be extended if needed.

    Parameters
    ----------
    role_id : int
        The primary key of the table.

    name : str
        The name of the role. Must be unique.

    rank : int
        The rank of the role. A role with a higher rank has more privileges. Used
        for comparing roles against one another. Two roles may have the same rank.

    description : str or None, default None
        A description of the role.

    updated_at : datetime or None
        The timestamp at which the role was last updated (UTC).

    updated_by : str or None
        The ID of the user that last updated the role.

    created_at : datetime
        The timestamp at which the role was created (UTC).
        Defaults to current timestamp.

    created_by : str or None
        The ID of the user that created the role.

    users : list[User]
        The users that have the role assigned.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = (
        'role_id',
        'name',
        'rank',
        'description',
        'updated_at',
        'updated_by',
        'created_at',
        'created_by',
    )

    __tablename__ = 'stp_role'

    role_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(unique=True)
    rank: Mapped[int]
    description: Mapped[str | None]
    users: Mapped[list['User']] = relationship(back_populates='role')

    @classmethod
    def create_viewer(cls) -> Self:
        r"""Create the VIEWER role."""

        return cls(
            role_id=1,
            name=UserRoleName.VIEWER,
            rank=1,
            description='A user that can only view data within an application.',
        )

    @classmethod
    def create_user(cls) -> Self:
        r"""Create the USER role, which is the default for a new user."""

        return cls(
            role_id=2,
            name=UserRoleName.USER,
            rank=2,
            description=(
                'The standard user with normal privileges. The default role for a new user.'
            ),
        )

    @classmethod
    def create_superuser(cls) -> Self:
        r"""Create the SUPERUSER role."""

        return cls(
            role_id=3,
            name=UserRoleName.SUPERUSER,
            rank=3,
            description=(
                'A user with higher privileges that can perform certain '
                'operations that a normal `USER` can not.'
            ),
        )

    @classmethod
    def create_admin(cls) -> Self:
        r"""Create the ADMIN role."""

        return cls(
            role_id=4,
            name=UserRoleName.ADMIN,
            rank=4,
            description=(
                'An admin has full access to everything. Only admin users may sign '
                'in to the admin page and manage the users of the application. '
                'An application should have at least one admin.'
            ),
        )


Index(f'{Role.__tablename__}_name_ix', Role.name)


class CustomRole(ModifiedAndCreatedColumnMixin, Base):
    r"""The custom roles of a user.

    Custom roles can be defined specifically for each application.
    A user may have none or multiple custom roles, which may grant
    application specific privileges or hide/expose specific pages.

    Parameters
    ----------
    role_id : int
        The primary key of the table.

    name : str
        The name of the role.

    rank : int
        The rank of the role. A role with a higher rank has more privileges. Used
        for comparing roles against one another. Two roles may have the same rank.

    description : str or None, default None
        A description of the role.

    updated_at : datetime or None
        The timestamp at which the custom role was last updated (UTC).

    updated_by : str or None
        The ID of the user that last updated the custom role.

    created_at : datetime
        The timestamp at which the custom role was created (UTC).
        Defaults to current timestamp.

    created_by : str or None
        The ID of the user that created the custom role.

    users : dict[str, User]
        A mapping of the users that have the custom role assigned.
        The key is the username of the user.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = (
        'role_id',
        'name',
        'rank',
        'description',
        'updated_at',
        'updated_by',
        'created_at',
        'created_by',
    )

    __tablename__ = 'stp_custom_role'

    role_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(unique=True)
    rank: Mapped[int]
    description: Mapped[Optional[str]]
    users: Mapped[dict[str, 'User']] = relationship(
        secondary='stp_user_custom_role_link',
        back_populates='custom_roles',
        collection_class=attribute_keyed_dict('username'),
    )


Index(f'{CustomRole.__tablename__}_name_ix', CustomRole.name)

# Association table for the many-to-many relationship between User and CustomRole.
user_custom_role_link = Table(
    'stp_user_custom_role_link',
    Base.metadata,
    Column('user_id', ForeignKey('stp_user.user_id', ondelete='CASCADE'), primary_key=True),
    Column('role_id', ForeignKey('stp_custom_role.role_id'), primary_key=True),
)


class User(ModifiedAndCreatedColumnMixin, Base):
    r"""The user table.

    Parameters
    ----------
    user_id : str
        The unique ID of the user. The primary key of the table.

    username : str
        The username of the user. It must be unique across all users.

    ad_username : Optional[str]
        The active directory username of the user.

    displayname : Optional[str]
        A descriptive name of the user that is easy to understand for a human.

    role_id : int
        The unique id of the role associated with the user.

    verified : bool, default False
        True if a user is verified and False otherwise. A user is verified when
        at least one verified email address is associated with the user.

    verified_at : Optional[datetime]
        The timestamp in UTC when the user was verified.

    disabled : bool, default False
        If False the user is enabled and if True the user is disabled.
        A disabled user is not able to register credentials or sign in.

    disabled_at : Optional[datetime]
        The timestamp in UTC when the user was disabled.

    updated_at : datetime or None
        The timestamp at which the user was last updated (UTC).

    updated_by : str or None
        The ID of the user that last updated the user.

    created_at : datetime
        The timestamp at which the user was created (UTC).
        Defaults to current timestamp.

    created_by : str or None
        The ID of the user that created the user.

    role : Role
        The role of the user.

    custom_roles : dict[str, CustomRole]
        The custom roles of the user. The role name is mapped to the :class:`CustomRole` model.

    emails : list[Email]
        The email addresses associated with the user.

    sign_ins : list[UserSignIn]
        Info about when the user has signed in to the application.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = (
        'user_id',
        'username',
        'ad_username',
        'displayname',
        'role_id',
        'verified',
        'verified_at',
        'disabled',
        'disabled_at',
        'updated_at',
        'updated_by',
        'created_at',
        'created_by',
    )

    __tablename__ = 'stp_user'

    user_id: Mapped[str] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(unique=True, nullable=False)
    ad_username: Mapped[Optional[str]]
    displayname: Mapped[Optional[str]]
    role_id: Mapped[int] = mapped_column(ForeignKey(Role.role_id))
    verified: Mapped[bool] = mapped_column(default=False)
    verified_at: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP())
    disabled: Mapped[bool] = mapped_column(default=False)
    disabled_at: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP())
    role: Mapped[Role] = relationship(back_populates='users')
    custom_roles: Mapped[dict[str, CustomRole]] = relationship(
        secondary='stp_user_custom_role_link',
        collection_class=attribute_keyed_dict('name'),
        back_populates='users',
        passive_deletes=True,
    )
    emails: Mapped[list['Email']] = relationship(
        back_populates='user', cascade='delete, delete-orphan', order_by='Email.rank'
    )
    sign_ins: Mapped[list['UserSignIn']] = relationship(
        back_populates='user', cascade='delete, delete-orphan'
    )


Index(f'{User.__tablename__}_username_ix', User.username)
Index(f'{User.__tablename__}_ad_username_ix', User.ad_username)
Index(f'{User.__tablename__}_disabled_ix', User.disabled)


class Email(ModifiedAndCreatedColumnMixin, Base):
    r"""Email addresses of a user.

    Parameters
    ----------
    email_id : int
        The primary key of the table.

    user_id : str
        The unique ID of the user the email address belongs to.

    email : str
        An email address of a user. Must be unique.

    rank : int
        The rank of the email, where 1 defines the primary email, 2 the secondary
        and 3 the tertiary etc ... A user can only have one email of each rank.

    verified : bool, default False
        True if the email address is verified and False otherwise.

    verified_at : Optional[datetime]
        The timestamp in UTC when the email address was verified by the user.

    disabled : bool, default False
        If the email address is disabled or not.

    disabled_at : Optional[datetime]
        The timestamp in UTC when the email address was disabled.

    updated_at : datetime or None
        The timestamp at which the email was last updated (UTC).

    updated_by : str or None
        The ID of the user that last updated the email.

    created_at : datetime
        The timestamp at which the email was created (UTC).
        Defaults to current timestamp.

    created_by : str or None
        The ID of the user that created the email.

    user : User
        The user object the email address belongs to.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = (
        'email_id',
        'user_id',
        'email',
        'rank',
        'verified',
        'verified_at',
        'disabled',
        'disabled_at',
        'updated_at',
        'updated_by',
        'created_at',
        'created_by',
    )

    __tablename__ = 'stp_email'
    __table_args__ = (UniqueConstraint('user_id', 'rank', name=f'{__tablename__}_user_id_rank_iu'),)

    email_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[str] = mapped_column(ForeignKey(User.user_id, ondelete='CASCADE'))
    email: Mapped[str] = mapped_column(unique=True)
    rank: Mapped[int]
    verified: Mapped[bool] = mapped_column(default=False)
    verified_at: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP())
    disabled: Mapped[bool] = mapped_column(default=False)
    disabled_at: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP())
    user: Mapped['User'] = relationship(back_populates='emails')


Index(f'{Email.__tablename__}_email_ix', Email.email)
Index(f'{Email.__tablename__}_user_id_ix', Email.user_id)


class UserSignIn(Base):
    r"""A logging table of when a user has signed in to the application.

    Parameters
    ----------
    user_sign_in_id : int
        The primary key of the table.

    user_id : str
        The unique ID of the user that signed in to the application.

    sign_in_timestamp : datetime
        The timestamp in timezone UTC when the user signed in.

    success : bool
        True if the user was successfully signed in and False otherwise.

    origin : str
        The domain name or IP-address of the application the user signed in to.

    device : str
        The device the user signed in from. E.g. a web browser.

    country : str
        The country code of the country that the user signed in from. E.g. SE for Sweden.

    credential_nickname : str
        The nickname of the passkey that was used to sign in.

    credential_id : str
        The ID of the passkey credential used to sign in.

    sign_in_type : str
        The type of sign in method. E.g. 'passkey_signin'.

    rp_id : str or None
        The ID of the relaying party, which is the server that
        verifies the credentials during the sign in process.

    created_at : datetime
        The timestamp at which the user sign in record was created (UTC).
        Defaults to current timestamp.

    user : User
        The user object the sign in entry belongs to.
    """

    columns__repr__: ClassVar[tuple[str, ...]] = (
        'user_sign_in_id',
        'user_id',
        'sign_in_timestamp',
        'success',
        'origin',
        'device',
        'country',
        'credential_nickname',
        'credential_id',
        'sign_in_type',
        'rp_id',
        'created_at',
    )

    __tablename__ = 'stp_user_sign_in'

    user_sign_in_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[str] = mapped_column(ForeignKey(User.user_id, ondelete='CASCADE'))
    sign_in_timestamp: Mapped[datetime] = mapped_column(TIMESTAMP())
    success: Mapped[bool]
    origin: Mapped[str]
    device: Mapped[str]
    country: Mapped[str]
    credential_nickname: Mapped[str]
    credential_id: Mapped[str]
    sign_in_type: Mapped[str]
    rp_id: Mapped[Optional[str]]
    created_at: Mapped[datetime] = created_at_column
    user: Mapped['User'] = relationship(back_populates='sign_ins')


Index(f'{UserSignIn.__tablename__}_ix1', UserSignIn.user_id, UserSignIn.sign_in_timestamp)
Index(f'{UserSignIn.__tablename__}_ix2', UserSignIn.origin)



r"""The entry point of the sign in page."""

# Third party
import streamlit as st

# Local
from streamlit_passwordless.app.config import (
    APP_HOME_PAGE_URL,
    APP_ISSUES_PAGE_URL,
    MAINTAINER_INFO,
    setup,
)
from streamlit_passwordless.app.controllers.sign_in import controller

ABOUT = f"""\
Sign in to the Streamlit Passwordless admin web app.

{MAINTAINER_INFO}
"""


def sign_in_page() -> None:
    r"""Run the sign in page of the Streamlit Passwordless admin app."""

    st.set_page_config(
        page_title='Admin Sign in - Streamlit Passwordless',
        page_icon=':key:',
        layout='centered',
        initial_sidebar_state='collapsed',
        menu_items={
            'About': ABOUT,
            'Get Help': APP_HOME_PAGE_URL,
            'Report a bug': APP_ISSUES_PAGE_URL,
        },
    )
    client, session_factory, _ = setup()

    with session_factory() as session:
        controller(client=client, db_session=session)


if __name__ == '__main__' or __name__ == '__page__':
    sign_in_page()



r"""The views of the admin page."""

# Standard library
import logging
import time
from typing import Sequence

# Third party
import pandas as pd
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.components.admin import (
    create_user_button,
    refresh_users_button,
    user_role_multiselect,
    user_selectbox,
    user_state_radio_button,
)
from streamlit_passwordless.app.logic.admin import (
    filter_selectable_users,
    get_selectable_users_from_database,
)
from streamlit_passwordless.app.session_state import SK_SELECTED_USER
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient
from streamlit_passwordless.components import (
    BannerMessageType,
    bitwarden_register_form_existing_user,
    delete_user_button,
    display_banner_message,
    update_user_form,
)
from streamlit_passwordless.components.sign_out import sign_out_button
from streamlit_passwordless.models import User

logger = logging.getLogger(__name__)


def title() -> None:
    r"""Render the title view of the admin page."""

    st.title('Streamlit Passwordless Admin Console')
    st.divider()


def sidebar(user: User) -> None:
    r"""Render the sidebar of the admin page.

    Parameters
    ----------
    user : streamlit_passwordless.User
        The signed in admin user.
    """

    sign_in = user.sign_in
    if sign_in is None:
        return None

    username = user.username if (dn := user.displayname) is None else dn

    user_info = f"""\
- **User :** {username} ({user.role.name})
- **Signed in at :** {sign_in.sign_in_timestamp.strftime(r'%Y-%m-%d %H:%M:%S %Z')}
- **Passkey :** {sign_in.credential_nickname}
"""

    with st.sidebar:
        sign_out_button(user=user)
        st.markdown(user_info)


def statistics_view(nr_users_database: int, nr_users_bwp: int, nr_passkeys: int) -> None:
    r"""The statistics view of the admin page.

    Display the number of users and the number of registered passkeys.

    Parameters
    ----------
    nr_users_database : int
        The number of users in the database.

    nr_users_bwp : int
        The number of registered users in Bitwarden Passwordless.dev.

    nr_passkeys : int
        The number of registered passkeys in Bitwarden Passwordless.dev.
    """

    nr_local_users_col, nr_bwp_users_col, nr_passkeys_col = st.columns([0.3, 0.4, 0.3])
    with nr_local_users_col:
        st.metric(label='Users in Database', value=nr_users_database)
    with nr_bwp_users_col:
        st.metric(label='Users Bitwarden Passwordless', value=nr_users_bwp)
    with nr_passkeys_col:
        st.metric(label='Registered Passkeys', value=nr_passkeys)


@st.fragment
def manage_users_view(
    df: pd.DataFrame,
    roles: Sequence[db.models.Role],
    custom_roles: Sequence[db.models.CustomRole],
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    on_update_refresh_after_n_seconds: float = 1.5,
) -> None:
    r"""Manage the users of the application.

    The view allows to create, update and delete users.

    Parameters
    ----------
    df : pandas.DataFrame
        The selectable users to manage.

    roles : Sequence[streamlit_passwordless.db.models.Role]
        The roles of the users to use for filtering the selectable users.

    custom_roles : Sequence[db.models.CustomRole],
        The custom roles that can be assigned to a user.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    on_update_refresh_after_n_seconds : float, default 1.5
        The number of seconds to wait before refreshing the page after a
        user has been successfully updated. Allows the user to have time
        to see the success banner before the page refreshes.
    """

    db_user = st.session_state.get(SK_SELECTED_USER)

    filter_col, space_col, button_col, statistics_col = st.columns(
        [0.3, 0.1, 0.3, 0.3],
        vertical_alignment='bottom',
        gap='small',
    )
    with filter_col:
        left_col, right_col = st.columns([0.4, 0.6])
        with left_col:
            selected_user_state = user_state_radio_button(
                default_disabled=False if db_user is None else db_user.disabled
            )
        with right_col:
            selected_roles = user_role_multiselect(roles=roles)

    with space_col:
        pass

    with button_col:
        left_col, right_col = st.columns(2)
        with left_col:
            refresh_users_button(db_session=db_session)
        with right_col:
            create_user_button(db_session=db_session)

    with statistics_col:
        statistics_view(nr_users_database=df.shape[0], nr_users_bwp=0, nr_passkeys=0)

    if df.shape[0] > 0:
        df = filter_selectable_users(
            df=df, disabled=selected_user_state, roles={r.role_id for r in selected_roles}
        )

    if db_user is None:
        preselected_user = None
    else:
        try:
            preselected_user = df.index.get_loc(db_user.user_id)
        except KeyError:  # If the user role was updated to a role that is not in the filter.
            preselected_user = None

    selected_user_id = user_selectbox(df=df, preselected_user=preselected_user)

    st.divider()  # User and email

    if selected_user_id:
        db_user = db.get_user_by_user_id(
            session=db_session, user_id=selected_user_id, disabled=None
        )
        st.session_state[SK_SELECTED_USER] = db_user
    else:
        db_user = None

    left_col, right_col = st.columns([0.5, 0.5], vertical_alignment='top')
    user_is_updated = False
    with left_col:
        title_col, delete_button_col = st.columns([0.7, 0.3], vertical_alignment='bottom')

        with title_col:
            st.markdown('### Update User')
        with delete_button_col:
            clicked, success = delete_user_button(
                client=client, db_session=db_session, user=db_user
            )
            user_is_deleted = clicked and success

        if selected_user_id is None:
            emails = []
            display_banner_message(
                message='Select a user to update!', message_type=BannerMessageType.INFO
            )
        elif db_user is None:
            emails = []
            if not user_is_deleted:
                display_banner_message(
                    message='Selected user not found in database!',
                    message_type=BannerMessageType.ERROR,
                )
        else:
            db_user, user_is_updated = update_user_form(
                db_session=db_session, user=db_user, roles=roles, custom_roles=custom_roles
            )
            st.session_state[SK_SELECTED_USER] = db_user
            st.write(db_user)
            emails = db_user.emails

    with right_col:
        st.metric(label='Emails', value=len(emails))
        st.write(emails)

    st.divider()  # Passkeys

    left_col, right_col = st.columns([0.5, 0.5], vertical_alignment='top')
    with left_col:
        bitwarden_register_form_existing_user(
            client=client, db_session=db_session, user=db_user, get_current_user=False
        )
    with right_col:
        st.metric(label='Passkeys', value=3)

    if (user_is_updated or user_is_deleted) and db_user is not None:
        logger.info(
            f'Successfully {"updated" if user_is_updated else "deleted"} user: {db_user.user_id}!'
        )
        get_selectable_users_from_database(db_session=db_session)
        time.sleep(on_update_refresh_after_n_seconds)
        st.rerun()



r"""Database operations on the user table."""

# Standard library
import logging
from typing import Literal, Sequence, overload

# Third party
import pandas as pd
from sqlalchemy import and_, select
from sqlalchemy.exc import SQLAlchemyError

# Local
from streamlit_passwordless import exceptions
from streamlit_passwordless.database.crud.custom_role import get_custom_roles
from streamlit_passwordless.models import User

from .. import models
from ..core import Session
from ..core import commit as db_commit

logger = logging.getLogger(__name__)


@overload
def get_all_users(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[False] = False,
    index_col: str = 'user_id',
) -> Sequence[models.User]: ...


@overload
def get_all_users(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[True] = True,
    index_col: str = 'user_id',
) -> pd.DataFrame: ...


def get_all_users(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: bool = False,
    index_col: str = 'user_id',
) -> pd.DataFrame | Sequence[models.User]:
    r"""Get all users from the database.

    Parameters
    ----------
    session : Session
        An active database session.

    skip : int, default 0
        The number of users to skip from the beginning of the result set.
        The users are order by their username.

    limit : int or None, default None
        Limit the result to desired number of users. If None all users are returned.

    as_df : bool, default False
        Return the users as a :class:`pandas.DataFrame` instead of a sequence of user models.

    index_col : str, default 'user_id'
        The column from the user table to use as the index column
        of the :class:`pandas.DataFrame` if `as_df` is True.

    Returns
    -------
    pandas.DataFrame or Sequence[streamlit_passwordless.db.models.User]
        The users from the database.
    """

    query = select(models.User).order_by(models.User.username).offset(skip).limit(limit)

    if as_df:
        return pd.read_sql_query(sql=query, con=session.bind, index_col=index_col)  # type: ignore
    else:
        return session.scalars(query).all()


def get_user_by_username(
    session: Session, username: str, disabled: bool | None = False
) -> models.User | None:
    r"""Get a user by its username.

    Parameters
    ----------
    session : Session
        An active database session.

    username : str
        The username to filter by.

    disabled : bool or None, default False
        True if filtering for a disabled user and False for an enabled user.
        If None filtering by a disabled or enabled user is omitted.

    Returns
    -------
    streamlit_passwordless.db.models.User or None
        The user with `username` or None if a user with `username` was not found.
    """

    if disabled is None:
        where_clause = models.User.username == username
    else:
        where_clause = and_(models.User.username == username, models.User.disabled == disabled)

    query = select(models.User).where(where_clause)

    return session.scalars(query).one_or_none()


def get_user_by_user_id(
    session: Session, user_id: str, disabled: bool | None = False, is_verified: bool | None = None
) -> models.User | None:
    r"""Get a user by user_id.

    Parameters
    ----------
    session : Session
        An active database session.

    user_id : str
        The user_id to filter by.

    disabled : bool or None, default False
        True if filtering for a disabled user and False for an active user.
        If None filtering by a disabled or enabled user is omitted.

    is_verified : bool or None, default None
        If True filter by verified users and if False by non-verified users.
        If None filtering by verified users is omitted.

    Returns
    -------
    streamlit_passwordless.db.models.User or None
        The user matching `user_id` or None if a user with `user_id` was not found.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the user from the database.
    """

    if disabled is None:
        where_clause = models.User.user_id == user_id
    else:
        where_clause = and_(models.User.user_id == user_id, models.User.disabled == disabled)

    query = select(models.User).where(where_clause)

    if is_verified is True:
        query = query.where(models.User.verified_at._is_not(None))
    elif is_verified is False:
        query = query.where(models.User.verified_at._is(None))
    else:
        pass

    try:
        return session.scalars(query).one_or_none()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError(
            f'Error loading user {user_id=} from database!', e=e
        ) from None


def create_user(
    session: Session,
    user: User,
    custom_roles: Sequence[models.CustomRole] | None = None,
    created_by_user_id: str | None = None,
    commit: bool = False,
) -> models.User:
    r"""Create a new user in the database.

    Parameters
    ----------
    session : Session
        An active database session.

    user : streamlit_passwordless.User
        The user to create.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole] or None, default None
        The custom roles from the active database `session` to associate with the user.
        If provided these roles will take precedence over the custom roles defined on
        `user` and avoids a database lookup since the custom roles already exist in the `session`.

    created_by_user_id : str or None, default None
        The ID of the user that is creating the new user.

    commit : bool, default False
        True if the added user should be committed after being added to the session and False
        to commit later. Note that the returned `db_user` object will be in a expired state
        if committing and will be re-loaded from the database upon next access.

    Returns
    -------
    db_user : streamlit_passwordless.db.models.User
        The database model of the created user.

    Raises
    ------
    streamlit_passwordless.DatabaseCreateUserError
        If an error occurs while saving the user to the database or if the
        role of the user does not exist in the database.
    """

    if (role_id := user.role.role_id) is None:
        error_msg = (
            f'Cannot create user "{user.username}" because user.role.role_id is not specified!'
        )
        logger.error(error_msg)
        raise exceptions.DatabaseCreateUserError(error_msg)

    db_user = models.User(
        user_id=user.user_id,
        username=user.username,
        ad_username=user.ad_username,
        displayname=user.displayname,
        role_id=role_id,
        verified_at=user.verified_at,
        disabled=user.disabled,
        disabled_at=user.disabled_at,
        created_by=created_by_user_id,
    )

    if emails := user.emails:
        db_emails = [models.Email(**email.model_dump()) for email in emails]
        db_user.emails.extend(db_emails)
        session.add_all(db_emails)

    if custom_roles:
        db_user.custom_roles = {model.name: model for model in custom_roles}
    elif _custom_roles := user.custom_roles:
        role_ids = {role_id for cr in _custom_roles.values() if (role_id := cr.role_id) is not None}
        db_custom_roles = get_custom_roles(session=session, role_ids=role_ids)
        db_user.custom_roles = {model.name: model for model in db_custom_roles}

    session.add(db_user)

    if commit:
        error_msg = (
            f'Unable to save user "{user.username}" to database! Check the logs for more details.'
        )
        try:
            db_commit(session=session, error_msg=error_msg)
        except exceptions.DatabaseError as e:
            logger.error(e.detailed_message)
            raise exceptions.DatabaseCreateUserError(error_msg, e=e) from None

    return db_user



r"""Unit tests for the config module of Streamlit Passwordless."""

# Standard library
from typing import ClassVar

# Third party
import pytest
from pydantic import AnyHttpUrl, ValidationError
from sqlalchemy import URL, make_url

# Local
from streamlit_passwordless import config, exceptions
from streamlit_passwordless.bitwarden_passwordless import BITWARDEN_PASSWORDLESS_API_URL
from streamlit_passwordless.config import (
    STP_BWP_PRIVATE_KEY,
    STP_BWP_PUBLIC_KEY,
    STP_BWP_URL,
    STP_DB_URL,
    STP_SECRETS_SECTION,
    ConfigManager,
)

# =============================================================================================
# Fixtures
# =============================================================================================


@pytest.fixture()
def mocked_streamlit_secrets(monkeypatch: pytest.MonkeyPatch) -> ConfigManager:
    r"""Mock `streamlit.secrets` with configuration for Streamlit Passwordless.

    Returns
    -------
    cm : streamlit_passwordless.ConfigManager
        A config manager with the expected configuration loaded from
        the mocked `streamlit.secrets`.
    """

    secrets = {
        STP_SECRETS_SECTION: {
            STP_BWP_PUBLIC_KEY: 'public_key',
            STP_BWP_PRIVATE_KEY: 'private_key',
            STP_BWP_URL: 'https://bwp_url.com',
            STP_DB_URL: 'sqlite:///mocked.db',
        },
        'some_other_section': {'test': 123},
    }
    exp_config = {
        'bwp_public_key': 'public_key',
        'bwp_private_key': 'private_key',
        'bwp_url': AnyHttpUrl('https://bwp_url.com'),  # type: ignore
        'db_url': make_url('sqlite:///mocked.db'),
    }
    monkeypatch.setattr(config.st, 'secrets', secrets)

    cm = ConfigManager.model_validate(exp_config)

    return cm


# =============================================================================================
# Tests
# =============================================================================================


class TestConfigManagerInit:
    r"""Tests for initializing `streamlit_passwordless.ConfigManager`.

    The ConfigManager handles the configuration for a Streamlit Passwordless application.
    """

    bwp_public_key: ClassVar[str] = 'bwp_public_key'
    bwp_private_key: ClassVar[str] = 'bwp_private_key'

    def test_init_with_default_values(self) -> None:
        r"""Test to initialize `ConfigManager` with the minimum config needed.

        The config values not specified should be filled in with the default values.
        """

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        cm = ConfigManager(bwp_public_key=self.bwp_public_key, bwp_private_key=self.bwp_private_key)

        # Verify
        # ===========================================================
        assert cm.bwp_public_key == self.bwp_public_key, 'bwp_public_key is incorrect!'
        assert cm.bwp_private_key == self.bwp_private_key, 'bwp_private_key is incorrect!'
        assert cm.bwp_url == BITWARDEN_PASSWORDLESS_API_URL, 'bwp_url is incorrect!'
        assert cm.db_url == make_url('sqlite:///streamlit_passwordless.db'), 'db_url is incorrect!'

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'bwp_url, exp_bwp_url',
        (
            pytest.param(
                'http://test.com', AnyHttpUrl('http://test.com'), id='str'  # type: ignore
            ),
            pytest.param(
                AnyHttpUrl('https://test.com'), AnyHttpUrl('https://test.com'), id='AnyHttpUrl'  # type: ignore
            ),
        ),
    )
    def test_bwp_url(self, bwp_url: str | AnyHttpUrl, exp_bwp_url: AnyHttpUrl) -> None:
        r"""Test the `bwp_url` parameter of `ConfigManager`."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        cm = ConfigManager(
            bwp_public_key=self.bwp_public_key,
            bwp_private_key=self.bwp_private_key,
            bwp_url=bwp_url,  # type: ignore
        )

        # Verify
        # ===========================================================
        assert cm.bwp_url == exp_bwp_url, 'bwp_url is incorrect!'

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'db_url, exp_db_url',
        (
            pytest.param(
                'sqlite:///test.db', make_url('sqlite:///test.db'), id='str'  # type: ignore
            ),
            pytest.param(make_url('sqlite://'), make_url('sqlite://'), id='URL'),  # type: ignore
        ),
    )
    def test_db_url(self, db_url: str | URL, exp_db_url: URL) -> None:
        r"""Test the `db_url` parameter of `ConfigManager`."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        cm = ConfigManager(
            bwp_public_key=self.bwp_public_key, bwp_private_key=self.bwp_private_key, db_url=db_url
        )

        # Verify
        # ===========================================================
        assert cm.db_url == exp_db_url, 'db_url is incorrect!'

        # Clean up - None
        # ===========================================================

    @pytest.mark.raises
    def test_invalid_bwp_url(self) -> None:
        r"""Test to provide an invalid http url for Bitwarden Passwordless.

        `streamlit_passwordless.StreamlitPasswordlessError` is expected to be raised.
        """

        # Setup
        # ===========================================================
        bwp_url = 'htt://invalid.com'

        # Exercise
        # ===========================================================
        with pytest.raises(exceptions.StreamlitPasswordlessError) as exc_info:
            ConfigManager(
                bwp_public_key=self.bwp_public_key,
                bwp_private_key=self.bwp_private_key,
                bwp_url=bwp_url,  # type: ignore
            )

        # Verify
        # ===========================================================
        error_msg = exc_info.exconly()
        print(error_msg)

        assert bwp_url in error_msg, 'bwp_url not in error_msg!'

        # Clean up - None
        # ===========================================================

    @pytest.mark.raises
    def test_invalid_db_url(self) -> None:
        r"""Test to provide an invalid database url.

        `streamlit_passwordless.StreamlitPasswordlessError` is expected to be raised.
        """

        # Setup
        # ===========================================================
        db_url = 'sqlite::///invalid.db'

        # Exercise
        # ===========================================================
        with pytest.raises(exceptions.StreamlitPasswordlessError) as exc_info:
            ConfigManager(
                bwp_public_key=self.bwp_public_key,
                bwp_private_key=self.bwp_private_key,
                db_url=db_url,
            )

        # Verify
        # ===========================================================
        error_msg = exc_info.exconly()
        print(error_msg)

        assert 'DatabaseInvalidUrlError' in error_msg, 'DatabaseInvalidUrlError not in error_msg!'
        assert db_url in error_msg, 'db_url not in error_msg!'

        # Clean up - None
        # ===========================================================

    def test_from_aliases(self) -> None:
        r"""Test to initialize `ConfigManager` using parameter aliases."""

        # Setup
        # ===========================================================
        data = {
            STP_BWP_PUBLIC_KEY: 'public_key',
            STP_BWP_PRIVATE_KEY: 'private_key',
            STP_BWP_URL: 'https://bwp_url.com',
            STP_DB_URL: 'sqlite:///mocked.db',
        }
        exp_cm = ConfigManager(
            bwp_public_key='public_key',
            bwp_private_key='private_key',
            bwp_url='https://bwp_url.com',  # type: ignore
            db_url='sqlite:///mocked.db',
        )

        # Exercise
        # ===========================================================
        cm = ConfigManager.model_validate(data)

        # Verify
        # ===========================================================
        assert cm == exp_cm

        # Clean up - None
        # ===========================================================

    @pytest.mark.raises
    def test_is_immutable(self) -> None:
        r"""Test that `ConfigManager` is immutable.

        `pydantic.ValidationError` is expected to be raised when an attribute
        of `ConfigManager` is trying to be changed after the instance is created.
        """

        # Setup
        # ===========================================================
        cm = ConfigManager(bwp_public_key=self.bwp_public_key, bwp_private_key=self.bwp_private_key)

        # Exercise
        # ===========================================================
        with pytest.raises(ValidationError) as exc_info:
            cm.bwp_public_key = ''  # type: ignore

        # Verify
        # ===========================================================
        error_msg = exc_info.exconly()
        print(error_msg)

        assert 'bwp_public_key' in error_msg, 'bwp_public_key not in error_msg!'

        # Clean up - None
        # ===========================================================


class TestConfigManagerLoad:
    r"""Tests for the classmethod `ConfigManager.load`."""

    def test_from_streamlit_secrets(self, mocked_streamlit_secrets: ConfigManager) -> None:
        r"""Test to load the config from `streamlit.secrets`.

        All config keys are defined and no environment variables are specified.
        """

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        cm = ConfigManager.load()

        # Verify
        # ===========================================================
        assert cm == mocked_streamlit_secrets

        # Clean up - None
        # ===========================================================

    def test_from_streamlit_secrets_override_with_env_variables(
        self, mocked_streamlit_secrets: ConfigManager, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        r"""Test to load the config from `streamlit.secrets`.

        The config key `bwp_private_key` is overridden by a value from
        an environment variable.
        """

        # Setup
        # ===========================================================
        private_key_env_var = 'private_key_env_var'
        monkeypatch.setenv(STP_BWP_PRIVATE_KEY, private_key_env_var)
        cm_exp = ConfigManager(
            bwp_public_key=mocked_streamlit_secrets.bwp_public_key,
            bwp_private_key=private_key_env_var,
            bwp_url=mocked_streamlit_secrets.bwp_url,
            db_url=mocked_streamlit_secrets.db_url,
        )

        # Exercise
        # ===========================================================
        cm = ConfigManager.load()

        # Verify
        # ===========================================================
        assert cm == cm_exp

        # Clean up - None
        # ===========================================================

    def test_from_empty_streamlit_secrets_override_with_env_variables(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        r"""Test to load the config from an empty `streamlit.secrets`.

        The config keys `bwp_public_key` and `bwp_private_key` are
        fetched from environment variables.
        """

        # Setup
        # ===========================================================
        public_key_env_var = 'public_key_env_var'
        monkeypatch.setenv(STP_BWP_PUBLIC_KEY, public_key_env_var)

        private_key_env_var = 'private_key_env_var'
        monkeypatch.setenv(STP_BWP_PRIVATE_KEY, private_key_env_var)

        monkeypatch.setattr(config.st, 'secrets', {})

        cm_exp = ConfigManager(
            bwp_public_key=public_key_env_var, bwp_private_key=private_key_env_var
        )

        # Exercise
        # ===========================================================
        cm = ConfigManager.load()

        # Verify
        # ===========================================================
        assert cm == cm_exp

        # Clean up - None
        # ===========================================================

    @pytest.mark.raises
    def test_no_config_defined(self, monkeypatch: pytest.MonkeyPatch) -> None:
        r"""Test to load the config when no configuration is defined.

        `streamlit_passwordless.StreamlitPasswordlessError` is expected to be raised.
        """

        # Setup
        # ===========================================================
        monkeypatch.setattr(config.st, 'secrets', {})
        for env_var in (STP_BWP_PUBLIC_KEY, STP_BWP_PRIVATE_KEY, STP_BWP_URL, STP_DB_URL):
            monkeypatch.delenv(env_var, raising=False)

        # Exercise
        # ===========================================================
        with pytest.raises(exceptions.StreamlitPasswordlessError) as exc_info:
            ConfigManager.load()

        # Verify
        # ===========================================================
        error_msg = exc_info.exconly()
        print(error_msg)

        assert 'bwp_public_key' in error_msg, 'bwp_public_key not in error_msg!'
        assert 'bwp_private_key' in error_msg, 'bwp_private_key not in error_msg!'

        # Clean up - None
        # ===========================================================



r"""The configuration of Streamlit Passwordless."""

# Standard library
import logging
import os
from datetime import timedelta
from typing import Self

# Third party
import streamlit as st
from pydantic import AliasChoices, AnyHttpUrl, ConfigDict, Field, field_validator

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless import exceptions
from streamlit_passwordless.bitwarden_passwordless import BITWARDEN_PASSWORDLESS_API_URL
from streamlit_passwordless.models import BaseModel

logger = logging.getLogger(__name__)

# Config keys
STP_SECRETS_SECTION = 'streamlit-passwordless'
STP_BWP_PUBLIC_KEY = 'STP_BWP_PUBLIC_KEY'
STP_BWP_PRIVATE_KEY = 'STP_BWP_PRIVATE_KEY'
STP_BWP_URL = 'STP_BWP_URL'
STP_DB_URL = 'STP_DB_URL'


class ConfigManager(BaseModel):
    r"""Handles the configuration for running Streamlit Passwordless apps.

    Parameters
    ----------
    bwp_public_key : str
        The public key of Bitwarden Passwordless.

    bwp_private_key : str
        The private key of Bitwarden Passwordless.

    bwp_url : pydantic.AnyHttpUrl or str, default AnyHttpUrl('https://v4.passwordless.dev')
        The base url of the backend API of Bitwarden Passwordless. Specify this url
        if you are self-hosting Bitwarden Passwordless.

    db_url : str or sqlalchemy.URL, default 'sqlite:///streamlit_passwordless.db'
        The SQLAlchemy database url of the Streamlit Passwordless user database.
    """

    model_config = ConfigDict(frozen=True)

    bwp_public_key: str = Field(validation_alias=AliasChoices('bwp_public_key', STP_BWP_PUBLIC_KEY))
    bwp_private_key: str = Field(
        validation_alias=AliasChoices('bwp_private_key', STP_BWP_PRIVATE_KEY)
    )
    bwp_url: AnyHttpUrl = Field(
        default=BITWARDEN_PASSWORDLESS_API_URL,
        validation_alias=AliasChoices('bwp_url', STP_BWP_URL),
    )
    db_url: str | db.URL = Field(
        default='sqlite:///streamlit_passwordless.db',
        validate_default=True,
        validation_alias=AliasChoices('db_url', STP_DB_URL),
    )

    @field_validator('db_url')
    @classmethod
    def validate_url(cls, url: str | db.URL) -> db.URL:
        r"""Validate the database url."""

        try:
            return db.create_db_url(url)
        except exceptions.DatabaseInvalidUrlError as e:
            raise ValueError(f'{type(e).__name__} : {str(e)}') from None

    @classmethod
    def load(cls) -> Self:
        r"""Load the configuration for a Streamlit Passwordless application.

        The configuration can be loaded from environment variables or Streamlit secrets.
        Values from environment variables will override Streamlit secrets. The Streamlit
        secrets can be defined in "~/.streamlit/secrets.toml" or "./.streamlit/secrets.toml",
        where config keys in the second file will take precedence over the first.

        Returns
        -------
        streamlit_passwordless.ConfigManager
            The loaded configuration.

        Raises
        ------
        streamlit_passwordless.StreamlitPasswordlessError
            If the required configuration could not be found or was invalid.
        """

        config_dict = dict(st.secrets.get(STP_SECRETS_SECTION, {}))

        config_mapping = {
            'db_url': STP_DB_URL,
            'bwp_public_key': STP_BWP_PUBLIC_KEY,
            'bwp_private_key': STP_BWP_PRIVATE_KEY,
            'bwp_url': STP_BWP_URL,
        }

        config_dict |= {
            config_key: value
            for config_key, config_key_name in config_mapping.items()
            if (value := os.getenv(config_key_name)) is not None
        }

        return cls.model_validate(config_dict)


@st.cache_resource(ttl=timedelta(hours=6))
def load_config() -> ConfigManager:
    r"""Load the configuration for a Streamlit Passwordless application.

    The configuration can be loaded from environment variables or Streamlit secrets.
    Values from environment variables will override Streamlit secrets. The Streamlit
    secrets can be defined in "~/.streamlit/secrets.toml" or "./.streamlit/secrets.toml",
    where config keys in the second file will take precedence over the first.

    The loaded configuration is a cached resource for 6 hours.

    Returns
    -------
    streamlit_passwordless.ConfigManager
        The loaded configuration.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If the required configuration could not be found or was invalid.
    """

    return ConfigManager.load()



r"""The ID:s (keys) of components used for identification and session state management."""

# =====================================================================================
# Admin
# =====================================================================================

ADMIN_USER_ROLE_MULTISELECT = 'stp-admin-user-role-multiselect'
ADMIN_USER_STATE_RADIO_BUTTON = 'stp-admin-user-state-radio-button'
ADMIN_USER_SELECTBOX = 'stp-admin-user-selectbox'
ADMIN_CREATE_USER_BUTTON = 'stp-admin-create-user-button'
ADMIN_REFRESH_USERS_BUTTON = 'stp-admin-refresh-users-button'



r"""The entry point of the Streamlit Passwordless admin app."""

# Standard library
from pathlib import Path

# Third party
import streamlit as st

# Local
from streamlit_passwordless.app.config import Pages

APP_PATH = Path(__file__)


def main() -> None:
    r"""The page router for the Streamlit Passwordless admin app."""

    pages = [
        st.Page(page=Pages.ADMIN, title='Admin', default=True),
        st.Page(page=Pages.SIGN_IN, title='Sign in', url_path='sign_in'),
    ]
    page = st.navigation(pages, position='hidden')
    page.run()


if __name__ == '__main__':
    main()



r"""Database operations on the role table (:class:`streamlit_passwordless.db.models.Role`)."""

# Standard library
from typing import Literal, Sequence, overload

# Third party
import pandas as pd
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

# Local
from streamlit_passwordless import exceptions
from streamlit_passwordless.database.core import Session
from streamlit_passwordless.database.core import commit as db_commit
from streamlit_passwordless.database.models import Role
from streamlit_passwordless.database.schemas.role import RoleCreate


@overload
def get_all_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[False] = False,
    index_col: str = 'role_id',
) -> Sequence[Role]: ...


@overload
def get_all_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: Literal[True] = True,
    index_col: str = 'role_id',
) -> pd.DataFrame: ...


def get_all_roles(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: bool = False,
    index_col: str = 'role_id',
) -> pd.DataFrame | Sequence[Role]:
    r"""Get all roles from the database.

    Parameters
    ----------
    session : Session
        An active database session.

    skip : int, default 0
        The number of roles to skip from the beginning of the result set.
        The roles are order by their rank.

    limit : int or None, default None
        Limit the result to desired number of roles. If None all roles are returned.

    as_df : bool, default False
        Return the roles as a :class:`pandas.DataFrame` instead of a sequence of role models.

    index_col : str, default 'role_id'
        The column from the role table to use as the index column
        of the :class:`pandas.DataFrame` if `as_df` is True.

    Returns
    -------
    pandas.DataFrame or Sequence[streamlit_passwordless.db.models.Role]
        The roles from the database.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the roles from the database.
    """

    query = select(Role).order_by(Role.rank).offset(skip).limit(limit)

    try:
        if as_df:
            return pd.read_sql_query(sql=query, con=session.bind, index_col=index_col)  # type: ignore
        else:
            return session.scalars(query).all()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError('Error loading roles from database!', e=e) from None


def get_role_by_name(session: Session, name: str) -> Role | None:
    r"""Get a role by its unique name.

    Parameters
    ----------
    session : Session
        An active database session.

    name : str
        The name of the role to filter by.

    Returns
    -------
    streamlit_passwordless.db.models.Role or None
        The role matching `name` or None if a role with `name` was not found.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the role from the database.
    """

    query = select(Role).where(Role.name == name)
    try:
        return session.scalars(query).one_or_none()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError(
            f'Error loading role by {name=} from database!', e=e
        ) from None


def get_role_by_role_id(session: Session, role_id: int) -> Role | None:
    r"""Get a role by role_id.

    Parameters
    ----------
    session : Session
        An active database session.

    role_id : int
        The role_id to filter by.

    Returns
    -------
    streamlit_passwordless.db.models.Role or None
        The role matching `role_id` or None if a role with `role_id` was not found.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the role from the database.
    """

    query = select(Role).where(Role.role_id == role_id)
    try:
        return session.scalars(query).one_or_none()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError('Error loading role from database!', e=e) from None


def create_role(session: Session, role: RoleCreate, commit: bool = False) -> Role:
    r"""Create a new role in the database.

    Parameters
    ----------
    session : Session
        An active database session.

    role : streamlit_passwordless.db.RoleCreate
        The schema of the role to create.

    commit : bool, default False
        True if the added role should be committed after being added to the session and False
        to commit later. Note that the returned `db_role` object will be in a expired state
        if committing and will be re-loaded from the database upon next access.

    Returns
    -------
    db_role : streamlit_passwordless.db.models.Role
        The database model of the created role.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while saving the role to the database.

    streamlit_passwordless.DatabaseStatementError
        If there is an error with the executed SQL statement.
    """

    db_role = Role(**role.model_dump())
    session.add(db_role)

    if commit:
        error_msg = f'Unable to save role {role.name} to database! Check the logs for more details.'
        db_commit(session=session, error_msg=error_msg)

    return db_role


def create_default_roles(session: Session, commit: bool = False) -> tuple[Role, ...]:
    r"""Create the default roles in the database.

    Parameters
    ----------
    session : Session
        An active database session.

    commit : bool, default False
        if True the default roles are committed to the database after being added
        to the session. If False (the default) they are only added to the session.

    Returns
    -------
    roles : tuple[streamlit_passwordless.db.models.Role, ...]
        The created default roles.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while saving the roles to the database.

    streamlit_passwordless.DatabaseStatementError
        If there is an error with the executed SQL statement.
    """

    roles = (Role.create_viewer(), Role.create_user(), Role.create_superuser(), Role.create_admin())
    session.add_all(roles)

    if commit:
        error_msg = (
            f'Unable to save default roles: {", ".join(r.name for r in roles)} to the database!\n'
            'Check the logs for more details.'
        )
        db_commit(session=session, error_msg=error_msg)

    return roles



r"""The entry point of the admin page."""

# Third party
import streamlit as st

# Local
from streamlit_passwordless.app.config import (
    APP_HOME_PAGE_URL,
    APP_ISSUES_PAGE_URL,
    MAINTAINER_INFO,
    Pages,
    setup,
)
from streamlit_passwordless.app.controllers.admin import controller
from streamlit_passwordless.authorization import authorized
from streamlit_passwordless.components.config import SK_USER
from streamlit_passwordless.models import AdminRole

ABOUT = f"""\
The Streamlit Passwordless admin web app.

Manage the users of your application.

{MAINTAINER_INFO}
"""


def admin_page() -> None:
    r"""Run the admin page of the Streamlit Passwordless admin app."""

    st.set_page_config(
        page_title='Admin - Streamlit Passwordless',
        page_icon=':bulb:',
        layout='wide',
        menu_items={
            'About': ABOUT,
            'Get Help': APP_HOME_PAGE_URL,
            'Report a bug': APP_ISSUES_PAGE_URL,
        },
    )
    client, session_factory, _ = setup()

    user = st.session_state.get(SK_USER)
    if user is None:
        return

    with session_factory() as session:
        controller(client=client, db_session=session, user=user)


admin_page_authorized = authorized(redirect=Pages.SIGN_IN, role=AdminRole)(admin_page)

if __name__ == '__main__' or __name__ == '__page__':
    admin_page_authorized()



r"""The schemas for the role model."""

# Standard library

# Third party

# Local
from .core import SchemaBaseModel


class RoleCreate(SchemaBaseModel):
    r"""The schema for creating a new role.

    Parameters
    ----------
    name : str
        The name of the role.

    rank : int
        The rank of the role. A role with a higher rank has more privileges. Used
        for comparing roles against one another. Two roles may have the same rank.

    description : str or None, default None
        A description of the role.
    """

    name: str
    rank: int
    description: str | None = None



r"""Components related to the sign out process."""

# Third party
import streamlit as st

# Local
from streamlit_passwordless.authorization import sign_out
from streamlit_passwordless.models import User

from .config import SK_USER
from .core import ButtonType
from .ids import KEY_SIGN_OUT_BUTTON


def sign_out_button(
    user: User | None = None,
    label: str = 'Sign out',
    icon: str | None = None,
    button_type: ButtonType = 'primary',
    help: str | None = None,
    use_container_width: bool = False,
) -> tuple[bool, User | None]:
    r"""Render the sign out button.

    If clicked the user is signed out from the application.
    The button is disabled if a user is not signed in.

    Parameters
    ----------
    user : streamlit_passwordless.User or None, default None
        The user to sign out. If None the user to sign out is loaded from the
        session state using key :attr:`streamlit_passwordless.SK_USER`.

    label : str, default 'Sign out'
        The label of the sign out button. GitHub-flavored Markdown is supported.

    icon : str or None, default None
        An optional icon to display next to the `label` of the button.
        See the `icon` parameter of :func:`streamlit.button` for more info.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    help : str or None, default None
        An optional help text to display when hovering over the button.

    use_container_width : bool, default False
        If True the button will take upp the full width of the parent container
        and if False it will be sized to fit the content of `label` and `icon`.

    Returns
    -------
    clicked : bool
        True if the button was clicked and False otherwise.

    user : streamlit_passwordless.User or None
        The signed out user. None is returned if a user has not signed in yet.
    """

    user = st.session_state.get(SK_USER) if user is None else user

    clicked = st.button(
        label=label,
        key=KEY_SIGN_OUT_BUTTON,
        type=button_type,
        icon=icon,
        help=help,
        use_container_width=use_container_width,
        disabled=True if user is None else not user.is_authenticated,
        on_click=sign_out,
        kwargs={'user': user},
    )

    return clicked, user



r"""Metadata about the streamlit-passwordless package."""

# Standard library
from datetime import date

__versiontuple__ = (0, 16, 0)
r"""The version of streamlit-passwordless in a comparable form.
Adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`_
(MAJOR.MINOR.PATCH).
"""

__version__ = '0.16.0'
r"""The streamlit-passwordless version string."""

__releasedate__ = date(2025, 6, 27)
r"""The release date of the version specified in :data:`__versiontuple__`."""



r"""The update user form component and its callback functions."""

# Standard library
from datetime import datetime
from typing import Sequence
from zoneinfo import ZoneInfo

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db

from . import config, core, ids


def _validate_form(db_session: db.Session, current_username: str) -> None:
    r"""Validate the input fields of the update user form.

    Updates the following session state keys:
    - `streamlit_passwordless.SK_UPDATE_USER_FORM_IS_VALID` :
        True if the form validation passed and False otherwise.

    - `streamlit_passwordless.SK_UPDATE_USER_FORM_VALIDATION_ERRORS` :
        A dictionary mapping of the user form field name to its error message.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    current_username : str
        The current username of the user.
    """

    form_is_valid = True
    validation_errors = {}

    validation_error_field = core.FormField.USERNAME
    if username := st.session_state[ids.UPDATE_USER_FORM_USERNAME_TEXT_INPUT]:

        if (new_username := username.strip().casefold()) != current_username:
            db_user, error_msg = core.get_user_from_database(
                session=db_session, username=new_username
            )

            if error_msg:  # DatabaseError
                form_is_valid = False
                validation_errors[validation_error_field] = error_msg
            elif db_user is not None:
                form_is_valid = False
                validation_errors[validation_error_field] = f'User {username} already exists!'
            else:
                pass

    else:
        validation_errors[validation_error_field] = 'The username field is required!'

    st.session_state[config.SK_UPDATE_USER_FORM_VALIDATION_ERRORS] = validation_errors
    st.session_state[config.SK_UPDATE_USER_FORM_IS_VALID] = form_is_valid


def _updated_user(
    user: db.models.User,
    username: str,
    with_displayname: bool,
    displayname: str | None,
    with_ad_username: bool,
    ad_username: str | None,
    with_role: bool,
    role_id: int,
    with_custom_roles: bool,
    custom_roles: Sequence[db.models.CustomRole] | None,
    with_disabled: bool,
    disabled: bool,
    updated_by_user_id: str | None,
) -> bool:
    r"""Update the fields of the user.

    Parameters
    ----------
    user : streamlit_passwordless.db.User
        The user to update.

    username : str
        The new username of the user.

    with_displayname : bool, default True
        If True the displayname field is updatable.

    displayname : str or None
        The new displayname of the user.

    with_ad_username : bool, default False
        If True the ad_username field is updatable.

    ad_username : str or None
        The new active directory username of the user.

    with_role : bool, default True
        If True the role of the user is updatable.

    role_id : int
        The new role_id of the user.

    with_custom_roles : bool, default False
        If True the custom roles of the user are updatable.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole] or None
        The new custom roles of the user.

    with_disabled_toggle : bool, default True
        If True the enabled or disabled state of the user is updatable.

    disabled : bool
        The new state of the disabled field of the user.

    updated_by_user_id : str or None
        The ID of the user that is updating the `user` to update.

    Returns
    -------
    bool
        True if any field of `user` was updated and False otherwise.
    """

    if (new_username := username.strip()) != user.username:
        user.username = new_username
        username_updated = True
    else:
        username_updated = False

    if with_displayname:
        new_displayname = None if displayname is None else displayname.strip()
        displayname_updated = user.displayname != new_displayname

        if displayname_updated:
            user.displayname = new_displayname
    else:
        displayname_updated = False

    if with_ad_username:
        new_ad_username = None if ad_username is None else ad_username.strip()
        ad_username_updated = user.ad_username != new_ad_username

        if ad_username_updated:
            user.ad_username = new_ad_username
    else:
        ad_username_updated = False

    if with_role:
        role_updated = user.role_id != role_id

        if role_updated:
            user.role_id = role_id
    else:
        role_updated = False

    if with_custom_roles:
        new_custom_roles = {cr.name: cr for cr in custom_roles} if custom_roles else {}

        new_custom_roles_set = set(new_custom_roles.keys())
        old_custom_roles_set = set(user.custom_roles.keys())
        diff = set.union(
            new_custom_roles_set.difference(old_custom_roles_set),
            old_custom_roles_set.difference(new_custom_roles_set),
        )
        custom_roles_updated = True if diff else False

        if custom_roles_updated:
            user.custom_roles = new_custom_roles
    else:
        custom_roles_updated = False

    if with_disabled:
        disabled_updated = user.disabled != disabled

        if disabled_updated:
            user.disabled = disabled
            user.disabled_at = datetime.now(tz=ZoneInfo('UTC')) if disabled else None
    else:
        disabled_updated = False

    user.updated_by = updated_by_user_id

    return any(
        [
            username_updated,
            displayname_updated,
            ad_username_updated,
            role_updated,
            custom_roles_updated,
            disabled_updated,
        ]
    )


def update_user_form(
    db_session: db.Session,
    user: db.models.User,
    with_displayname: bool = True,
    with_ad_username: bool = True,
    with_role: bool = True,
    with_custom_roles: bool = True,
    with_disabled_toggle: bool = True,
    roles: Sequence[db.models.Role] | None = None,
    custom_roles: Sequence[db.models.CustomRole] | None = None,
    title: str = '#### Update user',
    border: bool = True,
    submit_button_label: str = 'Update User',
    submit_button_type: core.ButtonType = 'primary',
    clear_on_submit: bool = True,
    banner_container: core.BannerContainer | None = None,
    updated_by_user_id: str | None = None,
    username_label: str = 'Username',
    username_max_length: int | None = 50,
    username_placeholder: str | None = 'john.doe',
    username_help: str | None = 'A unique identifier for the user.',
    displayname_label: str = 'Displayname',
    displayname_max_length: int | None = 50,
    displayname_placeholder: str | None = 'John Doe',
    displayname_help: str | None = 'A descriptive name of the user.',
    ad_username_label: str = 'Active Directory Username',
    ad_username_max_length: int | None = 50,
    ad_username_placeholder: str | None = 'ad.john.doe',
    ad_username_help: str | None = 'The active directory username of the user.',
    role_label: str = 'Role',
    role_placeholder: str = 'Choose a role',
    role_help: str | None = 'The role of the user.',
    custom_roles_label: str = 'Custom Roles',
    custom_roles_placeholder: str = 'Choose a custom role',
    custom_roles_max_selections: int | None = None,
    custom_roles_help: str | None = 'The custom roles to associate with the user.',
    disabled_toggle_label: str = 'Disabled User',
    disabled_toggle_help: str | None = 'If the user should be created as disabled or enabled.',
) -> tuple[db.models.User, bool]:
    r"""Render the update user form.

    Update information about an existing user in the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    user : streamlit_passwordless.db.User
        The user to update.

    with_displayname : bool, default True
        If True the displayname field is added to the form to enable to
        edit the displayname of the user.

    with_ad_username : bool, default False
        If True the ad_username field is added to the form to enable to
        edit the active directory username of the user.

    with_role : bool, default True
        If True the role selectbox is added to the form to enable to change
        the role of the user.

    with_custom_roles : bool, default False
        If True the custom roles multiselectbox is added to the form to enable
        to change the custom roles of the user.

    with_disabled_toggle : bool, default True
        If True the disabled user toggle switch will be added to the form, which enables
        changing the state of the user between enabled and disabled.

    roles : Sequence[streamlit_passwordless.Role] or None, default None
        The available roles that can be assigned to the user. If None the defined
        roles in the database will be loaded and made available for selection.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole] or None, default None
        The available custom roles that can be assigned to the user. If None the defined
        custom roles in the database will be loaded and made available for selection.

    title : str, default '#### Update User'
        The title of the update user form. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    submit_button_label : str, default 'Create User'
        The label of the form submit button.

    submit_button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the submit button. Emulates the `type` parameter of :func:`streamlit.button`.

    clear_on_submit : bool, default True
        If True the form fields will be cleared when the submit form button is clicked.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the update user process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the form.

    updated_by_user_id : str or None, default None
        The ID of the user that is updating the `user` to update.
        If None the ID of the currently signed in user is used.

    Other Parameters
    ----------------
    username_label : str, default 'Username'
        The label of the username field.

    username_max_length : int or None, default 50
        The maximum allowed number of characters of the username field.
        If None the upper limit is removed.

    username_placeholder : str or None, default 'john.doe'
        The placeholder of the username field. If None the placeholder is removed.

    username_help : str or None, default 'A unique identifier for the user.'
        The help text to display for the username field. If None the help text is removed.

    displayname_label : str, default 'Displayname'
        The label of the displayname field.

    displayname_max_length : int or None, default 50
        The maximum allowed number of characters of the displayname field.
        If None the upper limit is removed.

    displayname_placeholder : str or None, default 'John Doe'
        The placeholder of the displayname field. If None the placeholder is removed.

    displayname_help : str or None, default 'A descriptive name of the user.'
        The help text to display for the displayname field. If None the help text is removed.

    ad_username_label : str, default 'Active Directory Username'
        The label of the ad_username field.

    ad_username_max_length : int or None, default 50
        The maximum allowed number of characters of the ad_username field.
        If None the upper limit is removed.

    ad_username_placeholder : str or None, default 'ad.john.doe'
        The placeholder of the ad_username field. If None the placeholder is removed.

    ad_username_help : str or None, default 'The active directory username of the user.'
        The help text to display for the ad_username field. If None the help text is removed.

    role_label : str, default 'Role'
        The label of the role selectbox.

    role_placeholder : str, default 'Choose a role'
        The placeholder of the role selectbox if a role is not preselected.

    role_help: str or None, default 'The role of the user.'
        The help text to display for the role selectbox. If None the help text is removed.

    custom_roles_label : str, default 'Custom Roles',
        The label of the custom roles multiselectbox.

    custom_roles_placeholder : str, default 'Choose a custom role'
        The placeholder of the custom roles multiselectbox if the user does not have
        any custom roles assigned. If None the placeholder is removed.

    custom_roles_max_selections : int or None, default None
        The maximum number of custom roles that can be selected in the custom roles multiselectbox.
        If None there is no upper limit.

    custom_roles_help : str or None, default 'The custom roles to associate with the user.'
        The help text to display for the custom roles multiselectbox.
        If None the help text is removed.

    disabled_toggle_label : str, default 'Disabled user'
        The label for the disabled user toggle switch.

    disabled_toggle_help : str or None, default 'If the user should be created as disabled or enabled.'
        The help text to display for the disabled toggle switch. If None the help text is removed.

    Returns
    -------
    user : streamlit_passwordless.db.models.User
        The updated user.

    is_updated : bool
        True if the user was updated and False otherwise.
    """

    banner_container = st.empty() if banner_container is None else banner_container
    banner_container_mapping = {}

    with st.form(key=ids.UPDATE_USER_FORM, clear_on_submit=clear_on_submit, border=border):
        st.markdown(title)

        username_error_banner = st.empty()
        banner_container_mapping[core.FormField.USERNAME] = username_error_banner

        username = st.text_input(
            label=username_label,
            value=user.username,
            placeholder=username_placeholder,
            max_chars=username_max_length,
            help=username_help,
            key=ids.UPDATE_USER_FORM_USERNAME_TEXT_INPUT,
        )

        if with_displayname:
            displayname = st.text_input(
                label=displayname_label,
                value=user.displayname,
                placeholder=displayname_placeholder,
                max_chars=displayname_max_length,
                help=displayname_help,
                key=ids.UPDATE_USER_FORM_DISPLAYNAME_TEXT_INPUT,
            )
        else:
            displayname = None

        if with_ad_username:
            ad_username = st.text_input(
                label=ad_username_label,
                value=user.ad_username,
                placeholder=ad_username_placeholder,
                max_chars=ad_username_max_length,
                help=ad_username_help,
                key=ids.UPDATE_USER_FORM_AD_USERNAME_TEXT_INPUT,
            )
        else:
            ad_username = None

        if with_role:
            if roles is None:
                db_roles, error_msg = core.get_all_roles_from_database(session=db_session)
                if error_msg:
                    core.display_banner_message(
                        message=error_msg,
                        message_type=core.BannerMessageType.ERROR,
                        container=st.empty(),
                    )
            else:
                # Load the cached roles into the current session without a database lookup.
                db_roles = tuple(db_session.merge(r, load=False) for r in roles)

            role = st.selectbox(
                label=role_label,
                options=db_roles if db_roles else tuple(),
                index=db_roles.index(user.role) if roles else 0,  # Role of the current user.
                placeholder=role_placeholder,
                format_func=lambda r: r.name,
                help=role_help,
                key=ids.UPDATE_USER_FORM_ROLE_SELECTBOX,
            )
            role_id = role.role_id
        else:
            role_id = 1

        if with_custom_roles:
            if custom_roles is None:
                st.write(f'Session loading custom roles : {db_session}')
                db_custom_roles, error_msg = core.get_all_custom_roles_from_database(
                    session=db_session
                )
                if error_msg:
                    core.display_banner_message(
                        message=error_msg,
                        message_type=core.BannerMessageType.ERROR,
                        container=st.empty(),
                    )
            else:
                db_custom_roles = tuple(db_session.merge(r, load=False) for r in custom_roles)

            # Widgets return a copy of the selected values.
            # Accessing the selected custom roles from the session state returns the same objects
            # as specified in the `options` parameter. This is important because if copies of
            # custom roles that already exist in the database session are assigned to the user
            # SQLAlchemy will raise an exception.
            st.multiselect(
                label=custom_roles_label,
                options=db_custom_roles,
                default=[role for role in user.custom_roles.values()],
                max_selections=custom_roles_max_selections,
                placeholder=custom_roles_placeholder,
                format_func=lambda r: r.name,
                help=custom_roles_help,
                key=ids.UPDATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX,
            )

        if with_disabled_toggle:
            disabled = st.toggle(
                label=disabled_toggle_label,
                value=user.disabled,
                help=disabled_toggle_help,
                key=ids.UPDATE_USER_FORM_DISABLED_TOGGLE,
            )
        else:
            disabled = False

        clicked = st.form_submit_button(
            label=submit_button_label,
            type=submit_button_type,
            on_click=_validate_form,
            kwargs={'db_session': db_session, 'current_username': user.username},
        )

        if not clicked:
            return user, False

        if not st.session_state.get(config.SK_UPDATE_USER_FORM_IS_VALID):
            validation_error_key = config.SK_UPDATE_USER_FORM_VALIDATION_ERRORS
            core.process_form_validation_errors(
                validation_errors=st.session_state[validation_error_key],
                banner_container_mapping=banner_container_mapping,  # type: ignore
                default_banner_container=banner_container,
            )
            st.session_state[validation_error_key] = {}

            return user, False

    if updated_by_user_id is None:
        signed_in_user = config.get_current_user()
        updated_by = signed_in_user.user_id if signed_in_user else None
    else:
        updated_by = updated_by_user_id

    is_updated = _updated_user(
        user=user,
        username=username,
        with_displayname=with_displayname,
        displayname=displayname,
        with_ad_username=with_ad_username,
        ad_username=ad_username,
        with_role=with_role,
        role_id=role_id,
        with_custom_roles=with_custom_roles,
        custom_roles=st.session_state.get(
            ids.UPDATE_USER_FORM_CUSTOM_ROLES_MULTISELECTBOX, tuple()
        ),
        with_disabled=with_disabled_toggle,
        disabled=disabled,
        updated_by_user_id=updated_by,
    )

    if is_updated:
        success, error_msg = core.update_user_in_database(session=db_session, user=user)
        if not success:
            core.display_banner_message(
                message=error_msg,
                message_type=core.BannerMessageType.ERROR,
                container=banner_container,
            )
            return user, False

        msg = f'Successfully updated user: {user.username}!'
        core.display_banner_message(
            message=msg, message_type=core.BannerMessageType.SUCCESS, container=banner_container
        )
    else:
        msg = f'Nothing has changed on user: {user.username}!'
        core.display_banner_message(
            message=msg, message_type=core.BannerMessageType.WARNING, container=banner_container
        )

    return user, is_updated



r"""The exception hierarchy of streamlit-passwordless."""

# Standard library
from typing import Any


class StreamlitPasswordlessError(Exception):
    r"""The base Exception of streamlit-passwordless.

    Parameters
    ----------
    message : str
        The error message.

    data : Any, default None
        Optional extra data to include in the exception.

    e : Exception or None, default None
        The parent exception that is being wrapped and is
        used to extract additional error information from.

    Attributes
    ----------
    message : str
        The error message of the `message` parameter.

    name : str
        The name of the exception.

    parent_exception : Exception or None
        The wrapped parent exception if provided.

    parent_exception_name : str or None
        The name with module path of `parent_exception`.

    parent_message : str or None
        The error message of the parent exception.

    parent_full_message : str or None
        The full error message of the parent exception.
    """

    def __init__(self, message: str, data: Any = None, e: Exception | None = None) -> None:
        self.message = message
        self.data = data
        self.name = self.__class__.__name__

        if e is not None:
            self.parent_exception: Exception | None = e
            self.parent_exception_name: str | None = f'{e.__module__}.{e.__class__.__name__}'
            self.parent_full_message: str | None = str(e)
        else:
            self.parent_exception = None
            self.parent_exception_name = None
            self.parent_full_message = None

        if getattr(self, 'parent_message', None) is None:
            self.parent_message = None if e is None else self._get_parent_error_message(e=e)

        super().__init__(message)

    @staticmethod
    def _get_parent_error_message(e: Exception) -> Any:
        r"""Get the error message of the parent exception."""

        try:
            return e.args[0]
        except IndexError:
            return None

    @property
    def displayable_message(self) -> str:
        r"""An error message that is safe to display to the user."""

        return self.message

    @property
    def detailed_message(self) -> str:
        r"""An error message with more details.

        It may contain more sensitive data that should not be displayed to the user.
        It is useful to log this message.
        """

        return (
            f'Message : {self.message}\n'
            f'Parent Exception : {self.parent_exception_name}\n'
            f'Parent Full Message :{self.parent_full_message}'
        )


class RegisterUserError(StreamlitPasswordlessError):
    r"""Raised for errors when registering a new user."""


class SignInTokenVerificationError(StreamlitPasswordlessError):
    r"""Raised for errors when the backend is verifying the sign in token."""


class DatabaseError(StreamlitPasswordlessError):
    r"""Base exception for database related errors.

    Attributes
    ----------
    db_api_exception_name : str or None
        The name with module path to the underlying database api exception
        that triggered the database error.
    """

    def __init__(self, message: str, e: Exception | None = None) -> None:
        if e is not None:
            self.db_api_exception_name = self._get_db_api_exception_name(e=e)
            self.parent_message = self._get_parent_error_message(e=e)
        else:
            self.db_api_exception_name = None

        super().__init__(message, e=e)

    @staticmethod
    def _get_db_api_exception_name(e: Exception) -> str | None:
        r"""Get the name of the exception raised from the underlying database api."""

        if (orig := getattr(e, 'orig', None)) is None:
            return None
        else:
            return f'{orig.__module__}.{orig.__class__.__name__}'

    @staticmethod
    def _get_parent_error_message(e: Exception) -> str | None:
        r"""Get the error message from the parent exception."""

        if (orig := getattr(e, 'orig', None)) is None:
            return orig
        else:
            return str(orig)

    @property
    def detailed_message(self) -> str:
        return (
            f'{self.name}({self.message})\n'
            f'Parent Exception : {self.parent_exception_name}\n'
            f'DB API Exception : {self.db_api_exception_name}\n'
            f'Parent Exception Message : {self.parent_message}'
        )


class DatabaseStatementError(DatabaseError):
    r"""Raised for errors related to the executed SQL statement.

    Attributes
    ----------
    statement : str or None
        The SQL statement that caused the error.

    params : str or None
        The parameters supplied to `statement`.
    """

    def __init__(self, message: str, e: Exception | None) -> None:
        if e is not None:
            self.statement = getattr(e, 'statement')
            self.params = getattr(e, 'params')
        else:
            self.statement = None
            self.params = None

        super().__init__(message, e=e)

    @property
    def detailed_message(self) -> str:
        return (
            f'{self.name}({self.message})\n'
            f'Parent Exception : {self.parent_exception_name}\n'
            f'DB API Exception : {self.db_api_exception_name}\n'
            f'Statement : {self.statement}\n'
            f'Params : {self.params}\n'
            f'Parent Exception Message : {self.parent_message}'
        )


class DatabaseCreateUserError(DatabaseError):
    r"""Raised for errors related to creating a new user in the database.

    Parameters
    ----------
    message : str
        The error message.
    """


class DatabaseInvalidUrlError(DatabaseError):
    r"""Raised for invalid SQLAlchemy database url:s.

    Parameters
    ----------
    message : str
        The error message.

    url : str
        The invalid database url that caused the error.
    """

    def __init__(self, message: str, url: str, e: Exception | None) -> None:
        self.url = url
        super().__init__(message, e=e)



r"""The client to use for interacting with the Bitwarden Passwordless backend API."""

# Standard library
import logging
from datetime import datetime, timedelta
from typing import Any, Literal, TypeAlias

# Third party
from passwordless import (
    Credential,
    DeleteUser,
    PasswordlessClient,
    PasswordlessClientBuilder,
    PasswordlessError,
    PasswordlessOptions,
    RegisterToken,
    VerifySignIn,
)
from pydantic import AnyHttpUrl, Field, PrivateAttr

# Local
from streamlit_passwordless import common, exceptions, models

BackendClient: TypeAlias = PasswordlessClient
PasskeyCredential: TypeAlias = Credential

BITWARDEN_PASSWORDLESS_API_URL: AnyHttpUrl = AnyHttpUrl('https://v4.passwordless.dev')  # type: ignore

logger = logging.getLogger(__name__)


class BitwardenRegisterConfig(models.BaseModel):
    r"""The available passkey configuration when registering a new user.

    See the `Bitwarden Passwordless`_ documentation for more info about the parameters.

    .. _ Bitwarden Passwordless: https://docs.passwordless.dev/guide/api.html#register-token

    Parameters
    ----------
    attestation : Literal['none', 'direct', 'indirect'], default 'none'
        WebAuthn attestation conveyance preference. 'direct' and 'indirect' are exclusive to the
        Enterprise plan of Bitwarden Passwordless. Trial & Pro plans are limited to 'none'.

    authenticator_type : Literal['any', 'platform', 'cross-platform'], default 'any'
        WebAuthn authenticator attachment modality. 'platform' refers to platform specific options
        such as Windows Hello, FaceID or TouchID, while 'cross-platform' means roaming devices such
        as security keys. 'any' (default) means any authenticator type is allowed.

    discoverable : bool, default True
        True allows the user to sign in without a username or alias by creating a
        client-side discoverable credential.

    user_verification : Literal['preferred', 'required', 'discouraged'], default 'preferred'
        Set the preference for how user verification (e.g. PIN code or biometrics) works when
        authenticating.

    validity : timedelta, default timedelta(seconds=120)
        When the registration token expires and becomes invalid defined as an offset
        from the start of the registration process.

    alias_hashing : bool, default True
        True means that aliases for a user are hashed before they are stored in the
        Bitwarden Passwordless database.
    """

    attestation: Literal['none', 'direct', 'indirect'] = 'none'
    authenticator_type: Literal['any', 'platform', 'cross-platform'] = 'any'
    discoverable: bool = True
    user_verification: Literal['preferred', 'required', 'discouraged'] = 'preferred'
    validity: timedelta = timedelta(seconds=120)
    alias_hashing: bool = True

    @property
    def expires_at(self) -> datetime:
        r"""The expiry time of the registration token in timezone UTC."""

        return common.get_current_datetime() + self.validity


class BitwardenPasswordlessClient(models.BaseModel):
    r"""The client for interacting with Bitwarden Passwordless.

    Parameters
    ----------
    public_key : str
        The public key of the Bitwarden Passwordless application.

    private_key : str
        The private key of the Bitwarden Passwordless application.

    url : pydantic.AnyHttpUrl or str, default 'https://v4.passwordless.dev'
        The base url of the Bitwarden Passwordless application.

    register_config : streamlit_passwordless.BitwardenRegisterConfig
        The passkey configuration when registering a new user.
        If not specified the default configuration is used.
    """

    public_key: str
    private_key: str
    url: AnyHttpUrl = BITWARDEN_PASSWORDLESS_API_URL
    register_config: BitwardenRegisterConfig = Field(default_factory=BitwardenRegisterConfig)
    _backend_client: BackendClient = PrivateAttr()

    def model_post_init(self, __context: Any) -> None:
        r"""Setup the Bitwarden Passwordless backend client."""

        try:
            options = PasswordlessOptions(api_secret=self.private_key, api_url=str(self.url))
            self._backend_client = PasswordlessClientBuilder(options=options).build()
        except PasswordlessError as e:
            error_msg = 'Could not build Bitwarden Passwordless backend client!'
            raise exceptions.StreamlitPasswordlessError(error_msg, e=e) from None

    def __hash__(self) -> int:
        return hash(self.private_key + self.public_key)

    def create_register_token(self, user: models.User, discoverable: bool | None = None) -> str:
        r"""Create a register token to use for registering a passkey with the user's device.

        Parameters
        ----------
        user : streamlit_passwordless.User
            The user to register.

        discoverable : bool or None, default None
            If True create a discoverable passkey and if False a non-discoverable passkey.
            If None the setting for discoverability from `self.register_config` is used.

        Returns
        -------
        str
            The register token.

        Raises
        ------
        streamlit_passwordless.RegisterUserError
            If an error occurs while trying to create the register token
            using the Bitwarden Passwordless backend API.
        """

        register_config = self.register_config
        input_register_token = RegisterToken(
            user_id=user.user_id,
            username=user.username,
            display_name=user.displayname,
            attestation=register_config.attestation,
            authenticator_type=register_config.authenticator_type,
            discoverable=register_config.discoverable if discoverable is None else discoverable,
            user_verification=register_config.user_verification,
            aliases=user.aliases,
            alias_hashing=register_config.alias_hashing,
            expires_at=register_config.expires_at,
        )

        try:
            registered_token = self._backend_client.register_token(
                register_token=input_register_token
            )
        except PasswordlessError as e:
            error_msg = f'Error creating register token! {str(e)}'
            data = {
                'input_register_config': input_register_token,
                'problem_details': e.problem_details,
            }
            raise exceptions.RegisterUserError(error_msg, data=data, e=e) from None

        return registered_token.token  # type: ignore

    def verify_sign_in(self, token: str) -> models.UserSignIn:
        r"""Verify the sign in token with the backend to complete the sign in process.

        Parameters
        ----------
        token : str
            The token to verify.

        Returns
        -------
        streamlit_passwordless.UserSignIn
            Details from Bitwarden Passwordless about the user that signed in.

        Raises
        ------
        streamlit_passwordless.SignInTokenVerificationError
            If the `token` token cannot be verified successfully.

        streamlit_passwordless.StreamlitPasswordlessError
            If an instance of `streamlit_passwordless.UserSignIn` cannot be
            successfully created.
        """

        try:
            verified_user = self._backend_client.sign_in(verify_sign_in=VerifySignIn(token=token))
        except PasswordlessError as e:
            error_msg = f'Error verifying the sign in token! {str(e)}'
            data = {'token': token, 'problem_details': e.problem_details}
            raise exceptions.SignInTokenVerificationError(error_msg, data=data, e=e) from None

        return models.UserSignIn.model_validate(verified_user)

    def get_credentials(self, user_id: str, origin: str | None = None) -> list[PasskeyCredential]:
        r"""Get the registered passkey credentials for a user.

        Parameters
        ----------
        user_id : str
            The unique ID of the user.

        origin : str or None, default None
            Filter the credentials by origin. Origin is the domain name a
            credential is registred to. If None all credentials are returned.

        Returns
        -------
        credentials : list[streamlit_passwordless.PasskeyCredential]
            The registered passkey credentials of the user.

        Raises
        ------
        streamlit_passwordless.StreamlitPasswordlessError
            If the credentials cannot be retrieved from Bitwarden Passwordless.
        """

        try:
            credentials: list[PasskeyCredential] = self._backend_client.get_credentials(
                user_id=user_id
            )
        except PasswordlessError as e:
            error_msg = f'Error retrieving credentials for {user_id=}!\nproblem_details: {e.problem_details}'
            data = {
                'user_id': user_id,
                'problem_details': e.problem_details,
            }
            raise exceptions.StreamlitPasswordlessError(error_msg, data=data) from None

        return [c for c in credentials if c.origin == origin] if origin else credentials

    def delete_user(self, user_id: str) -> None:
        r"""Delete a user from Bitwarden Passwordless.

        Parameters
        ----------
        user_id : str
            The unique ID of the user to delete.

        Raises
        ------
        streamlit_passwordless.StreamlitPasswordlessError
            If there is an error with deleting the user from Bitwarden Passwordless.
        """

        try:
            self._backend_client.delete_user(DeleteUser(user_id=user_id))
        except PasswordlessError as e:
            error_msg = f'Error deleting {user_id=}!\nproblem_details: {e.problem_details}'
            data = {
                'user_id': user_id,
                'problem_details': e.problem_details,
            }
            raise exceptions.StreamlitPasswordlessError(error_msg, data=data, e=e) from None
